#!/usr/bin/env python3

"""
test/unit_tests_d/ut_xmrseed: Monero mnemonic unit test for the MMGen suite
"""

altcoin_dep = True

from mmgen.util import msg,msg_r,ymsg

from ..include.common import cfg,qmsg,vmsg

class unit_test:

	vectors = ( # private keys are reduced
		(   '148d78d2aba7dbca5cd8f6abcfb0b3c009ffbdbea1ff373d50ed94d78286640e', # Monero repo
			'velvet lymph giddy number token physics poetry unquoted nibs useful sabotage limits benches ' +
			'lifestyle eden nitrogen anvil fewest avoid batch vials washing fences goat unquoted',
		),
		(   'e8164dda6d42bd1e261a3406b2038dcbddadbeefdeadbeefdeadbeefdeadbe0f',
			'viewpoint donuts ardent template unveil agile meant unafraid urgent athlete rustled mime azure ' +
			'jaded hawk baby jagged haystack baby jagged haystack ramped oncoming point template'
		),
		(   '6900dea9753f5c7ced87b53bdcfb109a8417bca6a2797a708194157b227fb60b',
			'criminal bamboo scamper gnaw limits womanly wrong tuition birth mundane donuts square cohesive ' +
			'dolphin titans narrate fuel saved wrap aloof magically mirror together update wrap'
		),
		(   '0000000000000000000000000000000000000000000000000000000000000001',
			'abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey abbey ' +
			'abbey abbey abbey abbey abbey bamboo jaws jerseys abbey'
		),
		(   '1c95988d7431ecd670cf7d73f45befc6feffffffffffffffffffffffffffff0f',
			'powder directed sayings enmity bacon vapidly entrance bumper noodles iguana sleepless nasty flying ' +
			'soil software foamy solved soggy foamy solved soggy jury yawning ankle solved'
		),
		(   '2c94988d7431ecd670cf7d73f45befc6feffffffffffffffffffffffffffff0f',
			'memoir apart olive enmity bacon vapidly entrance bumper noodles iguana sleepless nasty flying soil ' +
			'software foamy solved soggy foamy solved soggy jury yawning ankle foamy'
		),
		(   '4bb0288c9673b69fa68c2174851884abbaaedce6af48a03bbfd25e8cd0364102',
			'rated bicycle pheasants dejected pouch fizzle shipped rash citadel queen avatar sample muzzle mews ' +
			'jagged origin yeti dunes obtains godfather unbending pastry vortex washing citadel'
		),
		(   '4bb0288c9673b69fa68c2174851884abbaaedce6af48a03bbfd25e8cd0364100',
			'rated bicycle pheasants dejected pouch fizzle shipped rash citadel queen avatar sample muzzle mews ' +
			'jagged origin yeti dunes obtains godfather unbending kangaroo auctions audio citadel'
		),
		(   '1d95988d7431ecd670cf7d73f45befc6feffffffffffffffffffffffffffff0e',
			'pram distance scamper enmity bacon vapidly entrance bumper noodles iguana sleepless nasty flying ' +
			'soil software foamy solved soggy foamy solved soggy hashing mullet onboard solved'
		),
		(   'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff0f',
			'foamy solved soggy foamy solved soggy foamy solved soggy foamy solved soggy foamy solved soggy ' +
			'foamy solved soggy foamy solved soggy jury yawning ankle soggy'
		),
	)

	def run_test(self,name,ut):

		def test_fromhex(b):
			vmsg('')
			qmsg('Checking seed to mnemonic conversion:')
			for privhex,chk in self.vectors:
				vmsg(f'    {chk}')
				chk = tuple(chk.split())
				res = b.fromhex(privhex)
				if use_monero_python:
					mp_chk = tuple( wl.encode(privhex).split() )
					assert res == mp_chk, f'check failed:\nres: {res}\nchk: {chk}'
				assert res == chk, f'check failed:\nres: {res}\nchk: {chk}'

		def test_tohex(b):
			vmsg('')
			qmsg('Checking mnemonic to seed conversion:')
			for chk,words in self.vectors:
				vmsg(f'    {chk}')
				res = b.tohex( words.split() )
				if use_monero_python:
					mp_chk = wl.decode( words )
					assert res == mp_chk, f'check failed:\nres: {res}\nchk: {mp_chk}'
				assert res == chk, f'check failed:\nres: {res}\nchk: {chk}'

		msg_r('Testing xmrseed conversion routines...')
		qmsg('')

		from mmgen.xmrseed import xmrseed

		b = xmrseed()
		b.check_wordlist(cfg)

		try:
			from monero.wordlists.english import English
			wl = English()
		except ImportError:
			use_monero_python = False
			ymsg('Warning: unable to import monero-python, skipping external library checks')
		else:
			use_monero_python = True

		test_fromhex(b)
		test_tohex(b)

		vmsg('')
		qmsg('Checking error handling:')

		bad_chksum_mn = ('abbey ' * 21 + 'bamboo jaws jerseys donuts').split()
		bad_word_mn = "admire zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo".split()
		bad_seed = 'deadbeef'
		good_mn = self.vectors[0][1].split()
		good_hex = self.vectors[0][0]
		bad_len_mn = good_mn[:22]

		th = b.tohex
		fh = b.fromhex

		hse = 'HexadecimalStringError'
		sle = 'SeedLengthError'
		ase = 'AssertionError'
		mne = 'MnemonicError'

		bad_data = (
			('hex',               hse, 'not a hexadecimal',     lambda:fh('xx')),
			('seed len',          sle, 'invalid seed byte len', lambda:fh(bad_seed)),
			('mnemonic type',     ase, 'must be list',          lambda:th('string')),
			('pad arg (fromhex)', ase, "invalid 'pad' arg",     lambda:fh(good_hex,pad=23)),
			('pad arg (tohex)',   ase, "invalid 'pad' arg",     lambda:th(good_mn,pad=23)),
			('word',              mne, "not in Monero",         lambda:th(bad_word_mn)),
			('checksum',          mne, "checksum",              lambda:th(bad_chksum_mn)),
			('seed phrase len',   mne, "phrase len",            lambda:th(bad_len_mn)),
		)

		ut.process_bad_data(bad_data)

		vmsg('')
		msg('OK')

		return True



#!/usr/bin/env python3

"""
test.unit_tests_d.ut_mn_entry: Mnemonic user entry unit test for the MMGen suite
"""

from mmgen.util import msg,msg_r
from ..include.common import cfg,qmsg

class unit_test:

	vectors = {
		'mmgen':   {
			'usl': 10, 'sw': 3, 'lw': 12,
			'idx_minimal': ( # None: non-unique match. False: no match
				('a',         None),
				('aa',        False),
				('as',        None),
				('ask',       70),
				('afte',      None),
				('after',     None),
				('aftern',    20),
				('afternoon', 20),
				('afternoons',False),
				('g',         None),
				('gg',        False),
				('z',         False),
				('abi',       False),
				('abo',       None),
				('abl',       0),
				('able',      0),
				('abler',     False),
				('you',       None),
				('yout',      1625),
				('youth',     1625),
				('youths',    False),
			),
		},
		'xmrseed': { 'usl': 3, 'sw': 4, 'lw': 12 },
		'bip39':   { 'usl': 4, 'sw': 3, 'lw': 8 },
	}

	def run_test(self,name,ut):

		msg_r('Testing MnemonicEntry methods...')

		from mmgen.mn_entry import mn_entry

		msg_r('\nTesting computed wordlist constants...')
		for wl_id in self.vectors:
			for j,k in (('uniq_ss_len','usl'),('shortest_word','sw'),('longest_word','lw')):
				a = getattr(mn_entry( cfg, wl_id ),j)
				b = self.vectors[wl_id][k]
				assert a == b, f'{wl_id}:{j} {a} != {b}'
		msg('OK')

		msg_r('Testing idx()...')
		qmsg('')
		junk = 'a g z aa gg zz aaa ggg zzz aaaa gggg zzzz aaaaaaaaaaaaaa gggggggggggggg zzzzzzzzzzzzzz'
		for wl_id in self.vectors:
			m = mn_entry( cfg, wl_id )
			qmsg('Wordlist: '+wl_id)
			for entry_mode in ('full','short'):
				for a,word in enumerate(m.wl):
					b = m.idx(word,entry_mode)
					assert a == b, f'{a} != {b} ({word!r} - entry mode: {entry_mode!r})'
				a = None
				for word in junk.split():
					b = m.idx(word,entry_mode)
					assert a == b, f'{a} != {b} ({word!r} - entry mode: {entry_mode!r})'
			if 'idx_minimal' in self.vectors[wl_id]:
				for vec in self.vectors[wl_id]['idx_minimal']:
					chk = vec[1]
					b = m.idx(vec[0],'minimal')
					if chk is False:
						assert b is None, (b,None)
					elif chk is None:
						assert type(b) is tuple, (type(b),tuple)
					elif type(chk) is int:
						assert b == chk, (b,chk)
		msg('OK')

		return True



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/cmdtest.py: Command test runner for the MMGen wallet system
"""

def check_segwit_opts():
	for k,m in (('segwit','S'),('segwit_random','S'),('bech32','B')):
		if getattr(cfg,k) and m not in proto.mmtypes:
			die(1,f'--{k.replace("_","-")} option incompatible with {proto.cls_name}')

def create_shm_dir(data_dir,trash_dir):
	# Laggy flash media can cause pexpect to fail, so create a temporary directory
	# under '/dev/shm' and put datadir and tmpdirs here.
	import shutil
	from subprocess import run
	if sys.platform == 'win32':
		for tdir in (data_dir,trash_dir):
			try:
				os.listdir(tdir)
			except:
				pass
			else:
				try:
					shutil.rmtree(tdir)
				except: # we couldn't remove data dir - perhaps regtest daemon is running
					try:
						run(['python3',os.path.join('cmds','mmgen-regtest'),'stop'],check=True)
					except:
						die(4,f'Unable to remove {tdir!r}!')
					else:
						time.sleep(2)
						shutil.rmtree(tdir)
			os.mkdir(tdir,0o755)
		shm_dir = 'test'
	else:
		tdir,pfx = '/dev/shm','mmgen-test-'
		try:
			run(f'rm -rf {tdir}/{pfx}*',shell=True,check=True)
		except Exception as e:
			die(2,f'Unable to delete directory tree {tdir}/{pfx}* ({e.args[0]})')
		try:
			import tempfile
			shm_dir = str(tempfile.mkdtemp('',pfx,tdir))
		except Exception as e:
			die(2,f'Unable to create temporary directory in {tdir} ({e.args[0]})')

		dest = os.path.join(shm_dir,os.path.basename(trash_dir))
		os.mkdir(dest,0o755)

		run(f'rm -rf {trash_dir}',shell=True,check=True)
		os.symlink(dest,trash_dir)

		dest = os.path.join(shm_dir,os.path.basename(data_dir))
		shutil.move(data_dir,dest) # data_dir was created by Config()
		os.symlink(dest,data_dir)

	return shm_dir

import sys, os, time, asyncio

if sys.argv[-1] == 'clean':
	os.environ['MMGEN_TEST_SUITE'] = '1'
	repo_root = os.path.normpath(os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]),os.pardir)))
	os.chdir(repo_root)
	sys.path[0] = repo_root

	from shutil import rmtree
	from test.overlay import get_overlay_tree_dir
	overlay_tree_dir = get_overlay_tree_dir(repo_root)
	rmtree(overlay_tree_dir,ignore_errors=True)
	print(f'Removed {os.path.relpath(overlay_tree_dir)!r}')
else:
	# overlay must be set up before importing mmgen mods!
	try:
		from include.test_init import repo_root
	except ImportError:
		from test.include.test_init import repo_root

from mmgen.cfg import Config,gc
from mmgen.color import red,yellow,green,blue,cyan,gray,nocolor,init_color
from mmgen.util import msg, Msg, bmsg, die, suf, make_timestr

from test.include.common import (
	set_globals,
	cmdtest_py_log_fn,
	cmdtest_py_error_fn,
	mk_tmpdir,
	cleandir,
	iqmsg,
	omsg,
	omsg_r,
	ok,
	start_test_daemons,
	stop_test_daemons,
	init_coverage,
)

try:
	os.unlink(os.path.join(repo_root,cmdtest_py_error_fn))
except:
	pass

os.environ['MMGEN_QUIET'] = '0' # for this script and spawned scripts

opts_data = {
	'sets': [
		('list_current_cmd_groups',True,'list_cmd_groups',True),
		('demo',True,'exact_output',True),
		('demo',True,'buf_keypress',True),
		('demo',True,'pexpect_spawn',True),
	],
	'text': {
		'desc': 'High-level tests for the MMGen Wallet suite',
		'usage':'[options] [command [..command]] | [command_group[.command_subgroup][:command]]',
		'options': """
-h, --help           Print this help message
--, --longhelp       Print help message for long options (common options)
-a, --no-altcoin     Skip altcoin tests (WIP)
-A, --no-daemon-autostart Don't start and stop daemons automatically
-B, --bech32         Generate and use Bech32 addresses
-b, --buf-keypress   Use buffered keypresses as with real human input
                     (often required on slow systems, or under emulation)
-c, --print-cmdline  Print the command line of each spawned command
-C, --coverage       Produce code coverage info using trace module
-x, --debug-pexpect  Produce debugging output for pexpect calls
--, --demo           Add extra delay after each send to make input visible.
                     Implies --exact-output --pexpect-spawn --buf-keypress
-d, --deps-only      Run a command or command subgroup’s dependencies without
                     running the command or command group itself.
-D, --no-daemon-stop Don't stop auto-started daemons after running tests
-E, --direct-exec    Bypass pexpect and execute a command directly (for
                     debugging only)
-e, --exact-output   Show the exact output of the MMGen script(s) being run
-G, --exclude-groups=G Exclude the specified command groups (comma-separated)
-l, --list-cmds      List the test script’s available commands
-L, --list-cmd-groups List the test script’s command groups and subgroups
-g, --list-current-cmd-groups List command groups for current configuration
-n, --names          Display command names instead of descriptions
-N, --no-timings     Suppress display of timing information
-o, --log            Log commands to file {lf!r}
-O, --pexpect-spawn  Use pexpect.spawn instead of popen_spawn (much slower,
                     kut does real terminal emulation)
-p, --pause          Pause between tests, resuming on keypress
-P, --profile        Record the execution time of each script
-q, --quiet          Produce minimal output.  Suppress dependency info
-r, --resume=c       Resume at command 'c' after interrupted run
-R, --resume-after=c Same, but resume at command following 'c'
-t, --step           After resuming, execute one command and stop
-S, --skip-deps      Skip dependency checking for command
-u, --usr-random     Get random data interactively from user
-T, --pexpect-timeout=T Set the timeout for pexpect
-v, --verbose        Produce more verbose output
-W, --no-dw-delete   Don't remove default wallet from data dir after dw tests
                     are done
-X, --exit-after=C   Exit after command 'C'
-y, --segwit         Generate and use Segwit addresses
-Y, --segwit-random  Generate and use a random mix of Segwit and Legacy addrs
""",
		'notes': """

If no command is given, the whole test suite is run for the currently
specified coin (default BTC).

For traceback output and error file support, set the EXEC_WRAPPER_TRACEBACK
environment var
"""
	},
	'code': {
		'options': lambda proto,help_notes,s: s.format(
				lf = cmdtest_py_log_fn
			)
	}
}

# we need some opt values before running opts.init, so parse without initializing:
po = Config(opts_data=opts_data,parse_only=True)._parsed_opts

data_dir = Config.test_datadir

# step 1: delete data_dir symlink in ./test;
if not po.user_opts.get('skip_deps'):
	try:
		os.unlink(data_dir)
	except:
		pass

# step 2: opts.init will create new data_dir in ./test (if not po.user_opts['skip_deps'])
cfg = Config(opts_data=opts_data)

if cfg.no_altcoin and cfg.coin != 'BTC':
	die(1,f'--no-altcoin incompatible with --coin={cfg.coin}')

set_globals(cfg)

from test.cmdtest_py_d.common import ( # this must be loaded after set_globals()
	get_file_with_ext,
	confirm_continue
)

type(cfg)._reset_ok += (
	'no_daemon_autostart',
	'names',
	'no_timings',
	'exit_after',
	'resuming',
	'skipping_deps' )

logging = cfg.log or os.getenv('MMGEN_EXEC_WRAPPER')

cfg.resuming = any(k in po.user_opts for k in ('resume','resume_after'))
cfg.skipping_deps = cfg.resuming or 'skip_deps' in po.user_opts

cmd_args = cfg._args

if cfg.pexpect_spawn and sys.platform == 'win32':
	die(1,'--pexpect-spawn option not supported on Windows platform, exiting')

if cfg.daemon_id and cfg.daemon_id in cfg.blacklisted_daemons.split():
	die(1,f'cmdtest.py: daemon {cfg.daemon_id!r} blacklisted, exiting')

network_id = cfg.coin.lower() + ('_tn' if cfg.testnet else '')

proto = cfg._proto

# step 3: move data_dir to /dev/shm and symlink it back to ./test:
trash_dir = os.path.join('test','trash')
trash_dir2 = os.path.join('test','trash2')

if not cfg.skipping_deps:
	shm_dir = create_shm_dir(data_dir,trash_dir)

check_segwit_opts()

testing_segwit = cfg.segwit or cfg.segwit_random or cfg.bech32

if cfg.test_suite_deterministic:
	cfg.no_timings = True
	init_color(num_colors=0)
	os.environ['MMGEN_DISABLE_COLOR'] = '1' # for this script and spawned scripts

if cfg.profile:
	cfg.names = True

if cfg.exact_output:
	qmsg = qmsg_r = lambda s: None
else:
	qmsg = cfg._util.qmsg
	qmsg_r = cfg._util.qmsg_r

if cfg.skipping_deps:
	cfg.no_daemon_autostart = True

from test.cmdtest_py_d.cfg import cfgs,fixup_cfgs

for k in cfgs:
	cfgs[k]['tmpdir'] = os.path.join('test','tmp',str(k))

fixup_cfgs()

utils = {
#	'check_deps': 'check dependencies for specified command (WIP)', # TODO
	'clean':      'clean specified tmp dir(s) (specify by integer, no arg = all dirs)',
}

def list_cmds():

	def gen_output():

		gm = CmdGroupMgr()
		cw,d = 0,[]

		yield green('AVAILABLE COMMANDS:')

		for gname in gm.cmd_groups:
			tg = gm.gm_init_group(None,gname,None,None)
			desc = tg.__doc__.strip() if tg.__doc__ else type(tg).__name__
			d.append( (gname,desc,gm.cmd_list,gm.dpy_data) )
			cw = max(max(len(k) for k in gm.dpy_data),cw)

		for gname,gdesc,clist,dpdata in d:
			yield '\n'+green(f'{gname!r} - {gdesc}:')
			for cmd in clist:
				data = dpdata[cmd]
				yield '    {:{w}} - {}'.format(
					cmd,
					(data if isinstance(data,str) else data[1]),
					w = cw )

		w = max(map(len,utils))

		yield '\n'+green('AVAILABLE UTILITIES:')

		for cmd in sorted(utils):
			yield f'  {cmd:{w}} - {utils[cmd]}'

	from mmgen.ui import do_pager
	do_pager('\n'.join(gen_output()))

	sys.exit(0)

def do_between():
	if cfg.pause:
		confirm_continue()
	elif (cfg.verbose or cfg.exact_output) and not cfg.skipping_deps:
		sys.stderr.write('\n')

def list_tmpdirs():
	return {k:cfgs[k]['tmpdir'] for k in cfgs}

def clean(usr_dirs=None,clean_overlay=True):
	if cfg.skipping_deps:
		return
	all_dirs = list_tmpdirs()
	dirnums = map(int,(usr_dirs if usr_dirs is not None else all_dirs))
	dirlist = list(map(str,sorted(dirnums)))
	for d in dirlist:
		if d in all_dirs:
			cleandir(all_dirs[d])
		else:
			die(1,f'{d}: invalid directory number')
	if dirlist:
		iqmsg(green('Cleaned tmp director{} {}'.format(
			suf(dirlist,'ies'),
			' '.join(dirlist))
		))
	cleandir(data_dir)
	cleandir(trash_dir)
	cleandir(trash_dir2)
	iqmsg(green(f'Cleaned directories {data_dir!r} {trash_dir!r} {trash_dir2!r}'))

	if clean_overlay:
		cleandir(overlay_tree_dir)
		iqmsg(green(f'Cleaned directory {os.path.relpath(overlay_tree_dir)!r}'))

def create_tmp_dirs(shm_dir):
	if sys.platform == 'win32':
		for cfg in sorted(cfgs):
			mk_tmpdir(cfgs[cfg]['tmpdir'])
	else:
		os.makedirs( os.path.join('test','tmp'), mode=0o755, exist_ok=True )
		for cfg in sorted(cfgs):
			src = os.path.join(shm_dir,cfgs[cfg]['tmpdir'].split('/')[-1])
			mk_tmpdir(src)
			try:
				os.unlink(cfgs[cfg]['tmpdir'])
			except OSError as e:
				if e.errno != 2:
					raise
			finally:
				os.symlink(src,cfgs[cfg]['tmpdir'])

def set_restore_term_at_exit():
	import termios,atexit
	fd = sys.stdin.fileno()
	old = termios.tcgetattr(fd)
	def at_exit():
		termios.tcsetattr(fd, termios.TCSADRAIN, old)
	atexit.register(at_exit)

class CmdGroupMgr:

	dpy_data = None

	from test.cmdtest_py_d.cfg import cmd_groups_dfl,cmd_groups_extra

	cmd_groups = cmd_groups_dfl.copy()
	cmd_groups.update(cmd_groups_extra)

	@staticmethod
	def create_cmd_group(cls,sg_name=None):

		cmd_group_in = dict(cls.cmd_group_in)

		if sg_name and 'subgroup.' + sg_name not in cmd_group_in:
			die(1,f'{sg_name!r}: no such subgroup in test group {cls.__name__}')

		def add_entries(key,add_deps=True,added_subgroups=[]):

			if add_deps:
				for dep in cmd_group_in['subgroup.'+key]:
					for e in add_entries(dep):
						yield e

			assert isinstance(cls.cmd_subgroups[key][0],str), f'header for subgroup {key!r} missing!'

			if not key in added_subgroups:
				for e in cls.cmd_subgroups[key][1:]:
					yield e
				added_subgroups.append(key)

		def gen():
			for name,data in cls.cmd_group_in:
				if name.startswith('subgroup.'):
					sg_key = name.removeprefix('subgroup.')
					if sg_name in (None,sg_key):
						for e in add_entries(
								sg_key,
								add_deps = sg_name and not cfg.skipping_deps,
								added_subgroups = [sg_name] if cfg.deps_only else [] ):
							yield e
					if cfg.deps_only and sg_key == sg_name:
						return
				elif not cfg.skipping_deps:
					yield (name,data)

		return tuple(gen())

	def load_mod(self,gname,modname=None):
		clsname,kwargs = self.cmd_groups[gname]
		if modname is None and 'modname' in kwargs:
			modname = kwargs['modname']
		import importlib
		modpath = f'test.cmdtest_py_d.ct_{modname or gname}'
		return getattr(importlib.import_module(modpath),clsname)

	def create_group(self,gname,sg_name,full_data=False,modname=None,is3seed=False,add_dpy=False):
		"""
		Initializes the list 'cmd_list' and dict 'dpy_data' from module's cmd_group data.
		Alternatively, if called with 'add_dpy=True', updates 'dpy_data' from module data
		without touching 'cmd_list'
		"""

		cls = self.load_mod(gname,modname)
		cdata = []

		def get_shared_deps(cmdname,tmpdir_idx):
			"""
			shared_deps are "implied" dependencies for all cmds in cmd_group that don't appear in
			the cmd_group data or cmds' argument lists.  Supported only for 3seed tests at present.
			"""
			if not hasattr(cls,'shared_deps'):
				return []

			return [k for k,v in cfgs[str(tmpdir_idx)]['dep_generators'].items()
						if k in cls.shared_deps and v != cmdname]

		if not hasattr(cls,'cmd_group'):
			cls.cmd_group = self.create_cmd_group(cls,sg_name)

		for a,b in cls.cmd_group:
			if is3seed:
				for n,(i,j) in enumerate(zip(cls.tmpdir_nums,(128,192,256))):
					k = f'{a}_{n+1}'
					if hasattr(cls,'skip_cmds') and k in cls.skip_cmds:
						continue
					sdeps = get_shared_deps(k,i)
					if isinstance(b,str):
						cdata.append( (k, (i,f'{b} ({j}-bit)',[[[]+sdeps,i]])) )
					else:
						cdata.append( (k, (i,f'{b[1]} ({j}-bit)',[[b[0]+sdeps,i]])) )
			else:
				cdata.append( (a, b if full_data else (cls.tmpdir_nums[0],b,[[[],cls.tmpdir_nums[0]]])) )

		if add_dpy:
			self.dpy_data.update(dict(cdata))
		else:
			self.cmd_list = tuple(e[0] for e in cdata)
			self.dpy_data = dict(cdata)

		return cls

	def gm_init_group(self,trunner,gname,sg_name,spawn_prog):
		kwargs = self.cmd_groups[gname][1]
		cls = self.create_group(gname,sg_name,**kwargs)
		cls.group_name = gname
		return cls(trunner,cfgs,spawn_prog)

	def get_cls_by_gname(self,gname):
		return self.load_mod( gname, self.cmd_groups[gname][1].get('modname') )

	def list_cmd_groups(self):
		ginfo = []
		for gname in self.cmd_groups:
			ginfo.append(( gname, self.get_cls_by_gname(gname) ))

		if cfg.list_current_cmd_groups:
			exclude = (cfg.exclude_groups or '').split(',')
			ginfo = [g for g in ginfo
						if network_id in g[1].networks
							and not g[0] in exclude
							and g[0] in tuple(self.cmd_groups_dfl) + tuple(cmd_args) ]
			desc = 'CONFIGURED'
		else:
			desc = 'AVAILABLE'

		def gen_output():
			yield green(f'{desc} COMMAND GROUPS AND SUBGROUPS:')
			yield ''
			for name,cls in ginfo:
				yield '  {} - {}'.format(
					yellow(name.ljust(13)),
					(cls.__doc__.strip() if cls.__doc__ else cls.__name__) )
				if hasattr(cls,'cmd_subgroups'):
					subgroups = {k:v for k,v in cls.cmd_subgroups.items() if not k.startswith('_')}
					max_w = max(len(k) for k in subgroups)
					for k,v in subgroups.items():
						yield '    + {} · {}'.format( cyan(k.ljust(max_w+1)), v[0] )

		from mmgen.ui import do_pager
		do_pager('\n'.join(gen_output()))

		Msg( '\n' + ' '.join(e[0] for e in ginfo) )
		sys.exit(0)

	def find_cmd_in_groups(self,cmd,group=None):
		"""
		Search for a test command in specified group or all configured command groups
		and return it as a string.  Loads modules but alters no global variables.
		"""
		if group:
			if not group in [e[0] for e in self.cmd_groups]:
				die(1,f'{group!r}: unrecognized group')
			groups = [self.cmd_groups[group]]
		else:
			groups = self.cmd_groups

		for gname in groups:
			cls = self.get_cls_by_gname(gname)

			if not hasattr(cls,'cmd_group'):
				cls.cmd_group = self.create_cmd_group(cls)

			if cmd in cls.cmd_group:             # first search the class
				return gname

			if cmd in dir(cls(None,None,None)):  # then a throwaway instance
				return gname # cmd might exist in more than one group - we'll go with the first

		return None

class CmdTestRunner:
	'cmdtest.py test runner'

	def __del__(self):
		if logging:
			self.log_fd.close()

	def __init__(self,data_dir,trash_dir):

		self.data_dir = data_dir
		self.trash_dir = trash_dir
		self.cmd_total = 0
		self.rebuild_list = {}
		self.gm = CmdGroupMgr()
		self.repo_root = repo_root
		self.skipped_warnings = []
		self.resume_cmd = None
		self.deps_only = None

		if logging:
			self.log_fd = open(cmdtest_py_log_fn,'a')
			self.log_fd.write(f'\nLog started: {make_timestr()} UTC\n')
			omsg(f'INFO → Logging to file {cmdtest_py_log_fn!r}')
		else:
			self.log_fd = None

		if cfg.coverage:
			coverdir,accfile = init_coverage()
			omsg(f'INFO → Writing coverage files to {coverdir!r}')
			self.pre_args = ['python3','-m','trace','--count','--coverdir='+coverdir,'--file='+accfile]
		else:
			self.pre_args = ['python3'] if sys.platform == 'win32' else []

		if cfg.pexpect_spawn:
			omsg('INFO → Using pexpect.spawn() for real terminal emulation')

		self.set_spawn_env()

	def set_spawn_env(self):

		self.spawn_env = dict(os.environ)
		self.spawn_env.update({
			'MMGEN_NO_LICENSE': '1',
			'MMGEN_BOGUS_SEND': '1',
			'MMGEN_TEST_SUITE_PEXPECT': '1',
			'EXEC_WRAPPER_DO_RUNTIME_MSG':'1',
			# if cmdtest.py itself is running under exec_wrapper, disable writing of traceback file for spawned script
			'EXEC_WRAPPER_TRACEBACK': '' if os.getenv('MMGEN_EXEC_WRAPPER') else '1',
		})

		if cfg.exact_output:
			from mmgen.term import get_terminal_size
			self.spawn_env['MMGEN_COLUMNS'] = str(get_terminal_size().width)
		else:
			self.spawn_env['MMGEN_COLUMNS'] = '120'

	def spawn_wrapper(
			self,
			cmd,
			args            = [],
			extra_desc      = '',
			no_output       = False,
			msg_only        = False,
			no_msg          = False,
			cmd_dir         = 'cmds',
			no_exec_wrapper = False,
			timeout         = None,
			pexpect_spawn   = None,
			direct_exec     = False,
			no_passthru_opts = False,
			spawn_env_override = None,
			exit_val        = None,
			env             = {}):

		self.exit_val = exit_val

		desc = self.tg.test_name if cfg.names else self.gm.dpy_data[self.tg.test_name][1]
		if extra_desc:
			desc += ' ' + extra_desc

		cmd_path = (
			cmd if cfg.system # cfg.system is broken for main test group with overlay tree
			else os.path.relpath(os.path.join(repo_root,cmd_dir,cmd)) )

		args = (
			self.pre_args +
			([] if no_exec_wrapper else ['scripts/exec_wrapper.py']) +
			[cmd_path] +
			([] if no_passthru_opts else self.passthru_opts) +
			self.tg.extra_spawn_args +
			args )

		try:
			qargs = ['{q}{}{q}'.format( a, q = "'" if ' ' in a else '' ) for a in args]
		except:
			msg(f'args: {args}')
			raise

		cmd_disp = ' '.join(qargs).replace('\\','/') # for mingw

		if logging:
			self.log_fd.write('[{}][{}:{}] {}\n'.format(
				proto.coin.lower(),
				self.tg.group_name,
				self.tg.test_name,
				cmd_disp))

		for i in args: # die only after writing log entry
			if not isinstance(i,str):
				die(2,'Error: missing input files in cmd line?:\nName: {}\nCmdline: {!r}'.format(
					self.tg.test_name,
					args ))

		if not no_msg:
			t_pfx = '' if cfg.no_timings else f'[{time.time() - self.start_time:08.2f}] '
			if cfg.verbose or cfg.print_cmdline or cfg.exact_output:
				omsg(green(f'{t_pfx}Testing: {desc}'))
				if not msg_only:
					clr1,clr2 = (nocolor,nocolor) if cfg.print_cmdline else (green,cyan)
					omsg(
						clr1('Executing: ') +
						clr2(repr(cmd_disp) if sys.platform == 'win32' else cmd_disp)
					)
			else:
				omsg_r('{a}Testing {b}: {c}'.format(
					a = t_pfx,
					b = desc,
					c = 'OK\n' if direct_exec or cfg.direct_exec else ''))

		if msg_only:
			return

		# NB: the `pexpect_spawn` arg enables hold_protect and send_delay while the corresponding cmdline
		# option does not.  For performance reasons, this is the desired behavior.  For full emulation of
		# the user experience with hold protect enabled, specify --buf-keypress or --demo.
		send_delay = 0.4 if pexpect_spawn is True or cfg.buf_keypress else None
		pexpect_spawn = pexpect_spawn if pexpect_spawn is not None else bool(cfg.pexpect_spawn)

		spawn_env = dict(spawn_env_override or self.tg.spawn_env)
		spawn_env.update({
			'MMGEN_HOLD_PROTECT_DISABLE': '' if send_delay else '1',
			'MMGEN_TEST_SUITE_POPEN_SPAWN': '' if pexpect_spawn else '1',
			'EXEC_WRAPPER_EXIT_VAL': '' if exit_val is None else str(exit_val),
		})
		spawn_env.update(env)

		from test.include.pexpect import MMGenPexpect
		return MMGenPexpect(
			args          = args,
			no_output     = no_output,
			spawn_env     = spawn_env,
			pexpect_spawn = pexpect_spawn,
			timeout       = timeout,
			send_delay    = send_delay,
			direct_exec   = direct_exec )

	def end_msg(self):
		t = int(time.time() - self.start_time)
		sys.stderr.write(green(
			f'{self.cmd_total} test{suf(self.cmd_total)} performed' +
			('\n' if cfg.no_timings else f'.  Elapsed time: {t//60:02d}:{t%60:02d}\n')
		))

	def init_group(self,gname,sg_name=None,cmd=None,quiet=False,do_clean=True):

		from test.cmdtest_py_d.cfg import cmd_groups_altcoin
		if cfg.no_altcoin and gname in cmd_groups_altcoin:
			omsg(gray(f'INFO → skipping test group {gname!r} (--no-altcoin)'))
			return None

		ct_cls = CmdGroupMgr().load_mod(gname)

		if sys.platform == 'win32' and ct_cls.win_skip:
			omsg(gray(f'INFO → skipping test {gname!r} (platform=win32)'))
			return None

		for k in ('segwit','segwit_random','bech32'):
			if getattr(cfg,k):
				segwit_opt = k
				break
		else:
			segwit_opt = None

		def gen_msg():
			yield ('{g}:{c}' if cmd else 'test group {g!r}').format(g=gname,c=cmd)
			if len(ct_cls.networks) != 1:
				yield f' for {proto.coin} {proto.network}'
			if segwit_opt:
				yield ' (--{})'.format( segwit_opt.replace('_','-') )

		m = ''.join(gen_msg())

		if segwit_opt and not ct_cls.segwit_opts_ok:
			iqmsg(gray(f'INFO → skipping {m}'))
			return None

		# 'networks = ()' means all networks allowed
		nws = [(e.split('_')[0],'testnet') if '_' in e else (e,'mainnet') for e in ct_cls.networks]
		if nws:
			coin = proto.coin.lower()
			nw = ('mainnet','testnet')[proto.testnet]
			for a,b in nws:
				if a == coin and b == nw:
					break
			else:
				iqmsg(gray(f'INFO → skipping {m} (network={nw})'))
				return None

		if do_clean:
			clean(ct_cls.tmpdir_nums,clean_overlay=False)

		if not quiet:
			bmsg('Executing ' + m)

		if (not self.daemon_started) and self.gm.get_cls_by_gname(gname).need_daemon:
			start_test_daemons(network_id,remove_datadir=True)
			self.daemon_started = True

		if hasattr(self,'tg'):
			del self.tg

		self.tg = self.gm.gm_init_group(self,gname,sg_name,self.spawn_wrapper)
		self.ct_clsname = type(self.tg).__name__

		# pass through opts from cmdline (po.user_opts)
		self.passthru_opts = ['--{}{}'.format(
				k.replace('_','-'),
				'' if cfg._uopts[k] is True else '=' + cfg._uopts[k]
			) for k in cfg._uopts if k in self.tg.base_passthru_opts + self.tg.passthru_opts]

		if cfg.resuming:
			rc = cfg.resume or cfg.resume_after
			offset = 1 if cfg.resume_after else 0
			self.resume_cmd = self.gm.cmd_list[self.gm.cmd_list.index(rc)+offset]
			omsg(f'INFO → Resuming at command {self.resume_cmd!r}')
			if cfg.step:
				cfg.exit_after = self.resume_cmd

		if cfg.exit_after and cfg.exit_after not in self.gm.cmd_list:
			die(1,f'{cfg.exit_after!r}: command not recognized')

		return self.tg

	def run_tests(self,cmd_args):
		self.start_time = time.time()
		self.daemon_started = False
		gname_save = None

		def parse_arg(arg):
			if '.' in arg:
				a,b = arg.split('.')
				return [a] + b.split(':') if ':' in b else [a,b,None]
			elif ':' in arg:
				a,b = arg.split(':')
				return [a,None,b]
			else:
				return [self.gm.find_cmd_in_groups(arg),None,arg]

		if cmd_args:
			for arg in cmd_args:
				if arg in self.gm.cmd_groups:
					if self.init_group(arg):
						for cmd in self.gm.cmd_list:
							self.check_needs_rerun(cmd,build=True)
							do_between()
				else:
					gname,sg_name,cmdname = parse_arg(arg)
					if gname:
						same_grp = gname == gname_save # same group as previous cmd: don't clean, suppress blue msg
						if self.init_group(gname,sg_name,cmdname,quiet=same_grp,do_clean=not same_grp):
							if cmdname:
								if cfg.deps_only:
									self.deps_only = cmdname
								try:
									self.check_needs_rerun(cmdname,build=True)
								except Exception as e: # allow calling of functions not in cmd_group
									if isinstance(e,KeyError) and e.args[0] == cmdname:
										ret = getattr(self.tg,cmdname)()
										if type(ret).__name__ == 'coroutine':
											asyncio.run(ret)
									else:
										raise
								do_between()
							else:
								for cmd in self.gm.cmd_list:
									self.check_needs_rerun(cmd,build=True)
									do_between()
							gname_save = gname
					else:
						die(1,f'{arg!r}: command not recognized')
		else:
			if cfg.exclude_groups:
				exclude = cfg.exclude_groups.split(',')
				for e in exclude:
					if e not in self.gm.cmd_groups_dfl:
						die(1,f'{e!r}: group not recognized')
			for gname in self.gm.cmd_groups_dfl:
				if cfg.exclude_groups and gname in exclude:
					continue
				if self.init_group(gname):
					for cmd in self.gm.cmd_list:
						self.check_needs_rerun(cmd,build=True)
						do_between()

		self.end_msg()

	def check_needs_rerun(
			self,
			cmd,
			build        = False,
			root         = True,
			force_delete = False,
			dpy          = False):

		self.tg.test_name = cmd

		if self.ct_clsname == 'CmdTestMain' and testing_segwit and cmd not in self.tg.segwit_do:
			return False

		rerun = root # force_delete is not passed to recursive call

		fns = []
		if force_delete or not root:
			# does cmd produce a needed dependency(ies)?
			ret = self.get_num_exts_for_cmd(cmd)
			if ret:
				for ext in ret[1]:
					fn = get_file_with_ext(cfgs[ret[0]]['tmpdir'],ext,delete=build)
					if fn:
						if force_delete:
							os.unlink(fn)
						else: fns.append(fn)
					else: rerun = True

		fdeps = self.generate_file_deps(cmd)
		cdeps = self.generate_cmd_deps(fdeps)

		for fn in fns:
			my_age = os.stat(fn).st_mtime
			for num,ext in fdeps:
				f = get_file_with_ext(cfgs[num]['tmpdir'],ext,delete=build)
				if f and os.stat(f).st_mtime > my_age:
					rerun = True

		for cdep in cdeps:
			if self.check_needs_rerun(cdep,build=build,root=False,dpy=cmd):
				rerun = True

		if build:
			if rerun:
				for fn in fns:
					if not root:
						os.unlink(fn)
				if not (dpy and cfg.skipping_deps):
					self.run_test(cmd)
				if not root:
					do_between()
		else:
			# If prog produces multiple files:
			if cmd not in self.rebuild_list or rerun is True:
				self.rebuild_list[cmd] = (rerun,fns[0] if fns else '') # FIX

		return rerun

	def run_test(self,cmd):

		if self.deps_only and cmd == self.deps_only:
			sys.exit(0)

		d = [(str(num),ext) for exts,num in self.gm.dpy_data[cmd][2] for ext in exts]

		# delete files depended on by this cmd
		arg_list = [get_file_with_ext(cfgs[num]['tmpdir'],ext) for num,ext in d]

		# remove shared_deps from arg list
		if hasattr(self.tg,'shared_deps'):
			arg_list = arg_list[:-len(self.tg.shared_deps)]

		if self.resume_cmd:
			if cmd != self.resume_cmd:
				return
			bmsg(f'Resuming at {self.resume_cmd!r}')
			self.resume_cmd = None
			cfg.skipping_deps = False
			cfg.resuming = False

		if cfg.profile:
			start = time.time()

		self.tg.test_name = cmd # NB: Do not remove, this needs to be set twice
		cdata = self.gm.dpy_data[cmd]
#		self.tg.test_dpydata = cdata
		self.tg.tmpdir_num = cdata[0]
#		self.tg.cfg = cfgs[str(cdata[0])] # will remove this eventually
		test_cfg = cfgs[str(cdata[0])]
		for k in (  'seed_len', 'seed_id',
					'wpasswd', 'kapasswd',
					'segwit', 'hash_preset',
					'bw_filename', 'bw_params', 'ref_bw_seed_id',
					'addr_idx_list', 'pass_idx_list' ):
			if k in test_cfg:
				setattr(self.tg,k,test_cfg[k])

		ret = getattr(self.tg,cmd)(*arg_list) # run the test
		if type(ret).__name__ == 'coroutine':
			ret = asyncio.run(ret)
		self.process_retval(cmd,ret)

		if cfg.profile:
			omsg('\r\033[50C{:.4f}'.format( time.time() - start ))

		if cmd == cfg.exit_after:
			sys.exit(0)

	def warn_skipped(self):
		if self.skipped_warnings:
			print(yellow('The following tests were skipped and may require attention:'))
			r = '-' * 72 + '\n'
			print(r+('\n'+r).join(self.skipped_warnings))

	def process_retval(self,cmd,ret):
		if type(ret).__name__ == 'MMGenPexpect':
			ret.ok(exit_val=self.exit_val)
			self.cmd_total += 1
		elif ret == 'ok':
			ok()
			self.cmd_total += 1
		elif ret == 'error':
			die(2,red(f'\nTest {self.tg.test_name!r} failed'))
		elif ret in ('skip','skip_msg','silent'):
			if ret == 'silent':
				self.cmd_total += 1
			elif ret == 'skip_msg':
				ok('SKIP')
		elif isinstance(ret,tuple) and ret[0] == 'skip_warn':
			self.skipped_warnings.append(
				'Test {!r} was skipped:\n  {}'.format(cmd,'\n  '.join(ret[1].split('\n'))))
		else:
			die(2,f'{cmd!r} returned {ret}')

	def check_deps(self,cmds): # TODO: broken
		if len(cmds) != 1:
			die(1,f'Usage: {gc.prog_name} check_deps <command>')

		cmd = cmds[0]

		if cmd not in self.gm.cmd_list:
			die(1,f'{cmd!r}: unrecognized command')

		if not cfg.quiet:
			omsg(f'Checking dependencies for {cmd!r}')

		self.check_needs_rerun(self.tg,cmd)

		w = max(map(len,self.rebuild_list)) + 1
		for cmd in self.rebuild_list:
			c = self.rebuild_list[cmd]
			m = 'Rebuild' if (c[0] and c[1]) else 'Build' if c[0] else 'OK'
			omsg('cmd {:<{w}} {}'.format( cmd+':', m, w=w ))

	def generate_file_deps(self,cmd):
		return [(str(n),e) for exts,n in self.gm.dpy_data[cmd][2] for e in exts]

	def generate_cmd_deps(self,fdeps):
		return [cfgs[str(n)]['dep_generators'][ext] for n,ext in fdeps]

	def get_num_exts_for_cmd(self,cmd):
		try:
			num = str(self.gm.dpy_data[cmd][0])
		except KeyError:
			qmsg_r(f'Missing dependency {cmd!r}')
			gname = self.gm.find_cmd_in_groups(cmd)
			if gname:
				kwargs = self.gm.cmd_groups[gname][1]
				kwargs.update({'add_dpy':True})
				self.gm.create_group(gname,None,**kwargs)
				num = str(self.gm.dpy_data[cmd][0])
				qmsg(f' found in group {gname!r}')
			else:
				qmsg(' not found in any command group!')
				raise
		dgl = cfgs[num]['dep_generators']
		if cmd in dgl.values():
			exts = [k for k in dgl if dgl[k] == cmd]
			return (num,exts)
		else:
			return None

if __name__ == '__main__':

	if not cfg.skipping_deps: # do this before list cmds exit, so we stay in sync with shm_dir
		create_tmp_dirs(shm_dir)

	if cfg.list_cmd_groups:
		CmdGroupMgr().list_cmd_groups()
	elif cfg.list_cmds:
		list_cmds()
	elif cmd_args and cmd_args[0] in utils:
		globals()[cmd_args[0]](*cmd_args[1:])
		sys.exit(0)

	if cfg.pause:
		set_restore_term_at_exit()

	from mmgen.exception import TestSuiteSpawnedScriptException

	try:
		tr = CmdTestRunner(data_dir,trash_dir)
		tr.run_tests(cmd_args)
		tr.warn_skipped()
		if tr.daemon_started:
			stop_test_daemons(network_id)
	except KeyboardInterrupt:
		if tr.daemon_started:
			stop_test_daemons(network_id)
		tr.warn_skipped()
		die(1,yellow('\ntest.py exiting at user request'))
	except TestSuiteSpawnedScriptException as e:
		# if spawned script is not running under exec_wrapper, output brief error msg:
		if os.getenv('MMGEN_EXEC_WRAPPER'):
			Msg(red(str(e)))
			Msg(blue('cmdtest.py: spawned script exited with error'))
		raise
	except Exception:
		# if cmdtest.py itself is running under exec_wrapper, re-raise so exec_wrapper can handle exception:
		if os.getenv('MMGEN_EXEC_WRAPPER') or not os.getenv('MMGEN_IGNORE_TEST_PY_EXCEPTION'):
			raise
		die(1,red('Test script exited with error'))



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.addresses: Bitcoin base protocol tracking wallet address list class
"""

from ....tw.addresses import TwAddresses
from ....tw.shared import TwLabel
from ....util import msg,msg_r
from ....obj import get_obj
from .rpc import BitcoinTwRPC

class BitcoinTwAddresses(TwAddresses,BitcoinTwRPC):

	has_age = True
	prompt_fs = """
Sort options: [a]mt, [A]ge, [M]mgen addr, [r]everse
Column options: toggle [D]ays/date/confs/block
Filters: show [E]mpty addrs, [u]sed addrs, all [L]abels
View/Print: pager [v]iew, [w]ide pager view, [p]rint{s}
Actions: [q]uit menu, r[e]draw, add [l]abel:
"""
	key_mappings = {
		'a':'s_amt',
		'A':'s_age',
		'M':'s_twmmid',
		'r':'s_reverse',
		'D':'d_days',
		'e':'d_redraw',
		'E':'d_showempty',
		'u':'d_showused',
		'L':'d_all_labels',
		'v':'a_view',
		'w':'a_view_detail',
		'p':'a_print_detail',
		'l':'i_comment_add' }

	async def get_rpc_data(self):

		msg_r('Getting unspent outputs...')
		addrs = await self.get_unspent_by_mmid(self.minconf)
		msg('done')

		amt0 = self.proto.coin_amt('0')
		self.total = sum((v['amt'] for v in addrs.values()), start=amt0)

		msg_r('Getting labels and associated addresses...')
		for e in await self.get_label_addr_pairs():
			if e.label and e.label.mmid not in addrs:
				addrs[e.label.mmid] = {
					'addr':   e.coinaddr,
					'amt':    amt0,
					'recvd':  amt0,
					'confs':  0,
					'lbl':    e.label}
		msg('done')

		msg_r('Getting received funds data...')
		# args: 1:minconf, 2:include_empty, 3:include_watchonly, 4:include_immature_coinbase (>=v23.0.0)
		for d in await self.rpc.call( 'listreceivedbylabel', 1, True, True ):
			label = get_obj( TwLabel, proto=self.proto, text=d['label'] )
			if label:
				assert label.mmid in addrs, f'{label.mmid!r} not found in addrlist!'
				addrs[label.mmid]['recvd'] = d['amount']
				addrs[label.mmid]['confs'] = d['confirmations']
		msg('done')

		return addrs



#!/usr/bin/env python3

import sys,os
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(os.path.dirname(sys.argv[0])))))
sys.path[0] = os.curdir

from mmgen.cfg import Config
cfg = Config()

from mmgen.term import init_term,get_term
init_term(cfg)
term = get_term()

if sys.argv[1] == 'echo':

	from mmgen.ui import line_input
	from mmgen.term import get_char_raw

	term.init(noecho=True)
	line_input( cfg, 'noecho> ' )
	get_char_raw()

	term.set('echo')
	line_input( cfg, 'echo> ' )

	term.set('noecho')
	line_input( cfg, 'noecho> ' )
	get_char_raw()

elif sys.argv[1] == 'cleanup':

	term.register_cleanup()

	import tty
	tty.setcbreak(term.stdin_fd)



#!/usr/bin/env python3

from Crypto.Cipher import AES
from Crypto.Util import Counter
ivec_len = 16
data = b'foobarbaz'
print('message:     ',data)
key = bytes.fromhex('deadbeef' * 8)
c = AES.new(key,AES.MODE_CTR,counter=Counter.new(ivec_len*8,initial_value=1))
enc_data = c.encrypt(data)
print('\npycrypto enc:',enc_data)
c = AES.new(key,AES.MODE_CTR,counter=Counter.new(ivec_len*8,initial_value=1))
dec_data = c.encrypt(enc_data)
print('pycrypto dec:', dec_data)

from cryptography.hazmat.primitives.ciphers import Cipher,algorithms,modes
from cryptography.hazmat.backends import default_backend

ctr_init_val = b'\x00' * 15 + b'\x01'
print('\nhazmat ctr_init_val:',ctr_init_val)
cipher = Cipher(algorithms.AES(key),modes.CTR(ctr_init_val),backend=default_backend())
encryptor = cipher.encryptor()
ct = encryptor.update(data)
#ct = encryptor.update(data) + encryptor.finalize()
print('\nhazmat enc:  ', ct)
assert ct == enc_data
decryptor = cipher.decryptor()
res = decryptor.update(ct)
#res = decryptor.update(ct) + decryptor.finalize()
print('hazmat dec:  ',res)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.completed: completed transaction class
"""

from .base import Base

class Completed(Base):
	"""
	signed or unsigned transaction with associated file
	"""
	filename_api = True

	def __init__(self,cfg,*args,filename=None,data=None,quiet_open=False,**kwargs):

		assert (filename or data) and not (filename and data), 'CompletedTX_chk1'

		super().__init__(cfg=cfg,*args,**kwargs)

		if data:
			self.__dict__ = data | {'twctl': self.twctl}
			self.name = type(self).__name__
		else:
			from .file import MMGenTxFile
			MMGenTxFile(self).parse(str(filename), quiet_open=quiet_open)

			self.check_serialized_integrity()

			# repeat with sign and send, because coin daemon could be restarted
			self.check_correct_chain()

			if self.check_sigs() != self.signed:
				from ..util import die
				die(1,'Transaction is {}signed!'.format('not ' if self.signed else ''))

			self.infile = filename

	@property
	def file(self):
		from .file import MMGenTxFile
		return MMGenTxFile(self)

	@classmethod
	def ext_to_cls(cls,ext,proto):
		"""
		see twctl:import_token()
		"""
		from .unsigned import Unsigned, AutomountUnsigned
		from .online import Sent, AutomountSent
		for cls in (Unsigned, AutomountUnsigned, Sent, AutomountSent):
			if ext == getattr(cls, 'ext'):
				return cls

		if proto.tokensym:
			from .online import OnlineSigned as Signed
			from .online import AutomountOnlineSigned as AutomountSigned
		else:
			from .signed import Signed, AutomountSigned
		for cls in (Signed, AutomountSigned):
			if ext == getattr(cls, 'ext'):
				return cls



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.bump: Ethereum transaction bump class
"""

from decimal import Decimal

from ....tx import bump as TxBase
from .completed import Completed,TokenCompleted
from .new import New,TokenNew

class Bump(Completed,New,TxBase.Bump):
	desc = 'fee-bumped transaction'

	@property
	def min_fee(self):
		return self.proto.coin_amt(self.fee * Decimal('1.101'))

	def bump_fee(self,idx,fee):
		self.txobj['gasPrice'] = self.fee_abs2rel(fee,to_unit='eth')

	async def get_nonce(self):
		return self.txobj['nonce']

class TokenBump(TokenCompleted,TokenNew,Bump):
	desc = 'fee-bumped transaction'

class AutomountBump(Bump):
	pass

class TokenAutomountBump(TokenBump):
	pass



#!/usr/bin/env python3

import sys,os,ast
#from pexpect.popen_spawn import PopenSpawn as spawn
from pexpect import spawn

# from mmgen.common import *
# from mmgen.obj import *
# cmd_args = opts.init(lambda: { 'desc': '', 'usage':'', 'options':'' })

p = spawn('./spawnprog.py foo bar',encoding='utf8')
ret = p.expect('hello: ')
print('got: {}'.format(ret))
print('before: [{}]'.format(p.before))
print('after: [{}]'.format(p.after))
print('groups: {}'.format(p.match.groups()))
ret = p.send('x')
ret = p.read()
print('got: ' + ret)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_py_d.ct_misc: Miscellaneous test groups for the cmdtest.py test suite
"""

import sys,os,re,time

from mmgen.util import ymsg

from ..include.common import cfg,start_test_daemons,stop_test_daemons,imsg
from .common import get_file_with_ext,dfl_words_file
from .ct_base import CmdTestBase
from .ct_main import CmdTestMain

class CmdTestDev(CmdTestBase):
	'developer scripts'
	networks = ('btc',)
	cmd_group = (
		('compute_file_chksum', 'scripts/compute-file-chksum.py'),
		('create_bip_hd_chain_params', 'scripts/create-bip-hd-chain-params.py'),
	)
	tmpdir_nums = [99]
	color = True

	def _spawn(self,script,args):
		return self.spawn(script,args,cmd_dir='.',no_exec_wrapper=True)

	def compute_file_chksum(self):
		t = self._spawn('scripts/compute-file-chksum.py', ['test/ref/25EFA3[2.34].testnet.rawtx'])
		t.expect('3df942')
		return t

	def create_bip_hd_chain_params(self):
		t = self._spawn('scripts/create-bip-hd-chain-params.py', ['test/ref/altcoin/slip44-mini.json'])
		t.expect('[defaults]')
		t.expect(r"secp.*0488ade4.*0488b21e.*0'\/0\/0",regex=True)
		t.expect('[bip-44]')
		t.expect('[bip-49]')
		t.match_expect_list(['0','BTC','x','m','P2SH','049d7878','049d7cb2','80','05','x','Bitcoin','1'])
		return t

class CmdTestMisc(CmdTestBase):
	'miscellaneous tests (RPC backends, xmrwallet_txview, term)'
	networks = ('btc',)
	tmpdir_nums = [99]
	passthru_opts = ('daemon_data_dir','rpc_port')
	cmd_group = (
		('rpc_backends',     'RPC backends'),
		('xmrwallet_txview', "'mmgen-xmrwallet' txview"),
		('xmrwallet_txlist', "'mmgen-xmrwallet' txlist"),
		('coin_daemon_info', "'examples/coin-daemon-info.py'"),
		('examples_bip_hd',  "'examples/bip_hd.py'"),
		('term_echo',        "term.set('echo')"),
		('term_cleanup',     'term.register_cleanup()'),
	)
	need_daemon = True
	color = True

	def rpc_backends(self):
		backends = cfg._autoset_opts['rpc_backend'][1]
		for b in backends:
			t = self.spawn_chk('mmgen-tool',[f'--rpc-backend={b}','daemon_version'],extra_desc=f'({b})')
		return t

	def xmrwallet_txview(self,op='txview'):
		if cfg.no_altcoin:
			return 'skip'
		files = get_file_with_ext('test/ref/monero','tx',no_dot=True,delete=False,return_list=True)
		t = self.spawn( 'mmgen-xmrwallet', [op] + files )
		res = t.read(strip_color=True)
		if op == 'txview':
			for s in (
				'Amount:    0.74 XMR',
				'Dest:      56VQ9M6k',
			):
				assert s in res, f'{s} not in {res}'
		elif op == 'txlist':
			assert re.search( '3EBD06-.*D94583-.*8BFA29-', res, re.DOTALL )
		return t

	def xmrwallet_txlist(self):
		return self.xmrwallet_txview(op='txlist')

	def examples_bip_hd(self):
		if cfg.no_altcoin:
			return 'skip'
		return self.spawn('examples/bip_hd.py',cmd_dir='.')

	def coin_daemon_info(self):
		if cfg.no_altcoin:
			coins = ['btc']
		else:
			coins = ['btc','ltc','eth']
			start_test_daemons('ltc','eth')
		t = self.spawn('examples/coin-daemon-info.py',coins,cmd_dir='.')
		for coin in coins:
			t.expect(coin.upper() + r'\s+mainnet\s+Up',regex=True)
		if cfg.pexpect_spawn:
			t.send('q')
		if not cfg.no_altcoin:
			stop_test_daemons('ltc','eth')
		return t

	def term_echo(self):

		def test_echo():
			t.expect('echo> ','foo\n')
			t.expect('foo')

		def test_noecho():
			t.expect('noecho> ','foo\n')
			import pexpect
			try:
				t.expect('foo')
			except pexpect.TIMEOUT:
				imsg('[input not echoed - OK]')
			t.send('x')

		if self.skip_for_win():
			return 'skip'

		t = self.spawn('test/misc/term_ni.py',['echo'],cmd_dir='.',pexpect_spawn=True,timeout=1)
		t.p.logfile = None
		t.p.logfile_read = sys.stdout if cfg.verbose or cfg.exact_output else None
		t.p.logfile_send = None

		test_noecho()
		test_echo()
		test_noecho()

		return t

	def term_cleanup(self):
		if self.skip_for_win():
			return 'skip'
		return self.spawn('test/misc/term_ni.py',['cleanup'],cmd_dir='.',pexpect_spawn=True)

class CmdTestHelp(CmdTestBase):
	'help, info and usage screens'
	networks = ('btc','ltc','bch','eth','xmr')
	passthru_opts = ('daemon_data_dir','rpc_port','coin','testnet')
	cmd_group = (
		('usage',                 (1,'usage message',[])),
		('version',               (1,'version message',[])),
		('license',               (1,'license message',[])),
		('helpscreens',           (1,'help screens',             [])),
		('longhelpscreens',       (1,'help screens (--longhelp)',[])),
		('show_hash_presets',     (1,'info screen (--show-hash-presets)',[])),
		('tool_help',             (1,"'mmgen-tool' usage screen",[])),
		('tool_cmd_usage',        (1,"'mmgen-tool' usage screen",[])),
		('test_help',             (1,"'cmdtest.py' help screens",[])),
		('tooltest_help',         (1,"'tooltest.py' help screens",[])),
	)

	def usage(self):
		t = self.spawn('mmgen-walletgen', ['foo'], exit_val=1)
		t.expect('USAGE: mmgen-walletgen')
		return t

	def version(self):
		t = self.spawn('mmgen-tool', ['--version'], exit_val=0)
		t.expect('MMGEN-TOOL version')
		return t

	def license(self):
		t = self.spawn(
			'mmgen-walletconv',
			['--stdout','--in-fmt=hex','--out-fmt=hex'],
			env = {'MMGEN_NO_LICENSE':''} )
		t.expect('to continue: ', 'w')
		t.expect('TERMS AND CONDITIONS') # start of GPL text
		if cfg.pexpect_spawn:
			t.send('G')
		t.expect('return for a fee.')    # end of GPL text
		if cfg.pexpect_spawn:
			t.send('q')
		t.expect('to continue: ', 'c')
		t.expect('data: ','beadcafe'*4 + '\n')
		t.expect('to confirm: ', 'YES\n')
		return t

	def spawn_chk_expect(self,*args,**kwargs):
		expect = kwargs.pop('expect')
		t = self.spawn(*args,**kwargs)
		t.expect(expect)
		if t.pexpect_spawn:
			time.sleep(0.4)
			t.send('q')
		t.read()
		t.ok()
		t.skip_ok = True
		return t

	def helpscreens(self,arg='--help',scripts=(),expect='USAGE:.*OPTIONS:',pager=True):

		scripts = list(scripts) or [s.replace('mmgen-','') for s in os.listdir('cmds')]

		if 'tx' not in self.proto.mmcaps:
			scripts = [s for s in scripts if not (s == 'regtest' or s.startswith('tx'))]

		if self.proto.coin not in ('BTC','XMR') and 'xmrwallet' in scripts:
			scripts.remove('xmrwallet')

		if sys.platform == 'win32' and 'autosign' in scripts:
			scripts.remove('autosign')

		for s in sorted(scripts):
			t = self.spawn(f'mmgen-{s}',[arg],extra_desc=f'(mmgen-{s})')
			t.expect(expect,regex=True)
			if pager and t.pexpect_spawn:
				time.sleep(0.2)
				t.send('q')
			t.read()
			t.ok()
			t.skip_ok = True

		return t

	def longhelpscreens(self):
		return self.helpscreens(arg='--longhelp',expect='USAGE:.*LONG OPTIONS:')

	def show_hash_presets(self):
		return self.helpscreens(
			arg = '--show-hash-presets',
			scripts = (
					'walletgen','walletconv','walletchk','passchg','subwalletgen',
					'addrgen','keygen','passgen',
					'txsign','txdo','txbump'),
			expect = 'Available parameters.*Preset',
			pager  = False )

	def tool_help(self):

		if os.getenv('PYTHONOPTIMIZE') == '2':
			ymsg('Skipping tool help with PYTHONOPTIMIZE=2 (no docstrings)')
			return 'skip'

		for arg in (
			'help',
			'usage',
		):
			t = self.spawn_chk_expect(
				'mmgen-tool',
				[arg],
				extra_desc = f'(mmgen-tool {arg})',
				expect = 'GENERAL USAGE' )
		return t

	def tool_cmd_usage(self):

		if os.getenv('PYTHONOPTIMIZE') == '2':
			ymsg('Skipping tool cmd usage with PYTHONOPTIMIZE=2 (no docstrings)')
			return 'skip'

		from mmgen.main_tool import mods

		for cmdlist in mods.values():
			for cmd in cmdlist:
				t = self.spawn_chk( 'mmgen-tool', ['help',cmd], extra_desc=f'({cmd})' )
		return t

	def test_help(self):
		for arg,expect in (
			('--help','USAGE'),
			('--list-cmds','AVAILABLE COMMANDS'),
			('--list-cmd-groups','AVAILABLE COMMAND GROUPS')
		):
			t = self.spawn_chk_expect(
				'cmdtest.py',
				[arg],
				cmd_dir = 'test',
				extra_desc = f'(cmdtest.py {arg})',
				expect = expect )
		return t

	def tooltest_help(self):
		for arg,expect in (
			('--list-cmds','Available commands'),
			('--testing-status','Testing status')
		):
			t = self.spawn_chk_expect(
				'tooltest.py',
				[arg],
				cmd_dir = 'test',
				extra_desc = f'(tooltest.py {arg})',
				expect = expect )
		return t

class CmdTestOutput(CmdTestBase):
	'screen output'
	networks = ('btc',)
	cmd_group = (
		('output_gr', (1,"Greek text", [])),
		('output_ru', (1,"Russian text", [])),
		('output_zh', (1,"Chinese text", [])),
		('output_jp', (1,"Japanese text", [])),
		('oneshot_warning', (1,"Oneshot warnings", [])),
		('oneshot_warning_term', (1,"Oneshot warnings (pexpect_spawn)", []))
	)
	color = True

	def screen_output(self,lang):
		return self.spawn('test/misc/utf8_output.py',[lang],cmd_dir='.')

	def output_gr(self):
		return self.screen_output('gr')
	def output_ru(self):
		return self.screen_output('ru')
	def output_zh(self):
		return self.screen_output('zh')
	def output_jp(self):
		return self.screen_output('jp')

	def oneshot_warning(self,pexpect_spawn=None):
		t = self.spawn('test/misc/oneshot_warning.py',cmd_dir='.',pexpect_spawn=pexpect_spawn)
		nl = '\r\n' if sys.platform == 'win32' or t.pexpect_spawn else '\n'
		for s in (
			f'pw{nl}wg1',
			'foo is experimental',
			'wg2', 'The bar command is dangerous',
			'wg3', 'baz variant alpha',
			'wg4', 'baz variant beta',
			'w1', 'foo variant alpha',
			'w2', 'foo variant beta',
			'w3', 'bar is experimental',
			'pw',
			"passphrase from file 'A'",
			"passphrase from file 'B'",
			f'wg1{nl}wg2{nl}wg3{nl}wg4{nl}w1{nl}w2{nl}w3',
			'pw',
			"passphrase from file 'A'",
			"passphrase from file 'B'",
			f'wg1{nl}wg2{nl}wg3{nl}wg4{nl}w1{nl}w2{nl}w3',
			'bottom',
		):
			t.expect(s)
		return t

	def oneshot_warning_term(self):
		if self.skip_for_win():
			return 'skip'
		return self.oneshot_warning(pexpect_spawn=True)

class CmdTestRefTX(CmdTestMain,CmdTestBase):
	'create a reference transaction file (administrative command)'
	segwit_opts_ok = False
	passthru_opts = ('daemon_data_dir','rpc_port','coin','testnet')
	tmpdir_nums = [31,32,33,34]
	need_daemon = True
	cmd_group = (
		('ref_tx_addrgen1', (31,'address generation (legacy)', [[[],1]])),
		('ref_tx_addrgen2', (32,'address generation (compressed)', [[[],1]])),
		('ref_tx_addrgen3', (33,'address generation (segwit)', [[[],1]])),
		('ref_tx_addrgen4', (34,'address generation (bech32)', [[[],1]])),
		('ref_tx_txcreate', (31,'transaction creation',
								([['addrs'],31],[['addrs'],32],[['addrs'],33],[['addrs'],34]))),
	)

	def __init__(self,trunner,cfgs,spawn):
		if cfgs:
			for n in self.tmpdir_nums:
				cfgs[str(n)].update({   'addr_idx_list': '1-2',
										'segwit': n in (33,34),
										'dep_generators': { 'addrs':'ref_tx_addrgen'+str(n)[-1] }})
		CmdTestMain.__init__(self,trunner,cfgs,spawn)

	def ref_tx_addrgen(self,atype):
		if atype not in self.proto.mmtypes:
			return
		return self.spawn('mmgen-addrgen',['--outdir='+self.tmpdir,'--type='+atype,dfl_words_file,'1-2'])

	def ref_tx_addrgen1(self):
		return self.ref_tx_addrgen(atype='L')
	def ref_tx_addrgen2(self):
		return self.ref_tx_addrgen(atype='C')
	def ref_tx_addrgen3(self):
		return self.ref_tx_addrgen(atype='S')
	def ref_tx_addrgen4(self):
		return self.ref_tx_addrgen(atype='B')

	def ref_tx_txcreate(self,f1,f2,f3,f4):
		sources = ['31','32']
		if 'S' in self.proto.mmtypes:
			sources += ['33']
		if 'B' in self.proto.mmtypes:
			sources += ['34']
		return self.txcreate_common(
				addrs_per_wallet = 2,
				sources          = sources,
				add_args         = ['--locktime=1320969600'],
				do_label         = True)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_py_d.ct_seedsplit: Seed split/join tests for the cmdtest.py test suite
"""

import os

from mmgen.wallet import get_wallet_cls
from mmgen.util import capfirst

from ..include.common import strip_ansi_escapes,cmp_or_die
from .common import get_file_with_ext
from .ct_base import CmdTestBase

ref_wf = 'test/ref/98831F3A.bip39'
ref_sid = '98831F3A'
wpasswd = 'abc'
sh1_passwd = 'xyz'
dfl_wcls = get_wallet_cls('mmgen')

class CmdTestSeedSplit(CmdTestBase):
	'splitting and joining seeds'
	networks = ('btc',)
	tmpdir_nums = [23]
	color = True
	cmd_group = (
		('ss_walletgen',                'wallet generation'),
		('ss_2way_A_dfl1',              '2-way seed split (share A)'),
		('ss_2way_B_dfl1',              '2-way seed split (share B)'),
		('ss_2way_join_dfl1',           '2-way seed join'),
		('ss_2way_A_dfl2',              "2-way seed split 'default' (share A)"),
		('ss_2way_B_dfl2',              "2-way seed split 'default' (share B)"),
		('ss_2way_join_dfl2',           "2-way seed join 'default'"),
		('ss_2way_A_alice',             "2-way seed split 'alice' (share A)"),
		('ss_2way_B_alice',             "2-way seed split 'alice' (share B)"),
		('ss_2way_join_alice',          "2-way seed join 'alice'"),
		('ss_2way_join_alice_mix',      "2-way seed join 'alice' (out of order)"),
		('ss_2way_A_dfl_master3',       '2-way seed split with master share #3 (share A)'),
		('ss_2way_B_dfl_master3',       '2-way seed split with master share #3 (share B)'),
		('ss_2way_join_dfl_master3',    '2-way seed join with master share #3'),
		('ss_2way_A_dfl_usw',           '2-way seed split of user-specified wallet (share A)'),
		('ss_2way_B_dfl_usw',           '2-way seed split of user-specified wallet (share B)'),
		('ss_2way_join_dfl_usw',        '2-way seed join of user-specified wallet'),
		('ss_3way_A_dfl',               '3-way seed split (share A)'),
		('ss_3way_B_dfl',               '3-way seed split (share B)'),
		('ss_3way_C_dfl',               '3-way seed split (share C)'),
		('ss_3way_join_dfl',            '3-way seed join'),
		('ss_3way_join_dfl_mix',        '3-way seed join (out of order)'),
		('ss_3way_A_foobar_master7',    "3-way seed split 'φυβαρ' with master share #7 (share A)"),
		('ss_3way_B_foobar_master7',    "3-way seed split 'φυβαρ' with master share #7 (share B)"),
		('ss_3way_C_foobar_master7',    "3-way seed split 'φυβαρ' with master share #7 (share C)"),
		('ss_3way_join_foobar_master7', "3-way seed join 'φυβαρ' with master share #7"),
		('ss_3way_join_foobar_master7_mix',"3-way seed join 'φυβαρ' with master share #7 (out of order)"),

		('ss_3way_join_dfl_bad_invocation',"bad invocation of 'mmgen-seedjoin' - --id-str with non-master join"),
		('ss_bad_invocation1',          "bad invocation of 'mmgen-seedsplit' - no arguments"),
		('ss_bad_invocation2',          "bad invocation of 'mmgen-seedsplit' - master share with split specifier"),
		('ss_bad_invocation3',          "bad invocation of 'mmgen-seedsplit' - nonexistent file"),
		('ss_bad_invocation4',          "bad invocation of 'mmgen-seedsplit' - invalid file extension"),
		('ss_bad_invocation5',          "bad invocation of 'mmgen-seedjoin' - no arguments"),
		('ss_bad_invocation6',          "bad invocation of 'mmgen-seedjoin' - one file argument"),
		('ss_bad_invocation7',          "bad invocation of 'mmgen-seedjoin' - invalid file extension"),
		('ss_bad_invocation8',          "bad invocation of 'mmgen-seedjoin' - nonexistent file"),
		('ss_bad_invocation9',          "bad invocation of 'mmgen-seedsplit' - bad specifier"),
		('ss_bad_invocation10',         "bad invocation of 'mmgen-seedsplit' - nonexistent file"),
		('ss_bad_invocation11',         "bad invocation of 'mmgen-seedsplit' - invalid file extension"),
	)

	def get_tmp_subdir(self,subdir):
		return os.path.join(self.tmpdir,subdir)

	def ss_walletgen(self):
		t = self.spawn('mmgen-walletgen', ['-r0','-p1'])
		t.passphrase_new('new '+dfl_wcls.desc,wpasswd)
		t.label()
		self.write_to_tmpfile('dfl.sid',strip_ansi_escapes(t.expect_getend('Seed ID: ')))
		t.expect('move it to the data directory? (Y/n): ','y')
		t.written_to_file(capfirst(dfl_wcls.desc))
		return t

	def ss_splt(self,tdir,ofmt,spec,add_args=[],wf=None,master=None):
		try:
			os.mkdir(self.get_tmp_subdir(tdir))
		except:
			pass
		t = self.spawn('mmgen-seedsplit',
				['-q','-d',self.get_tmp_subdir(tdir),'-r0','-o',ofmt]
				+ (['-L',(spec or 'label')] if ofmt == 'w' else [])
				+ add_args
				+ ([f'--master-share={master}'] if master else [])
				+ ([wf] if wf else [])
				+ ([spec] if spec else []))
		if not wf:
			t.passphrase(dfl_wcls.desc,wpasswd)
		if spec:
			from mmgen.seedsplit import SeedSplitSpecifier
			sss = SeedSplitSpecifier(spec)
			pat = rf'Processing .*\b{sss.idx}\b of \b{sss.count}\b of .* id .*‘{sss.id}’'
		else:
			pat = f'master share #{master}'
		t.expect(pat,regex=True)
		ocls = get_wallet_cls(fmt_code=ofmt)
		if ocls.enc:
			t.hash_preset('new '+ocls.desc,'1')
			t.passphrase_new('new '+ocls.desc,sh1_passwd)
			if ocls.type == 'incog_hidden':
				t.hincog_create(1234)
		t.written_to_file(capfirst(ocls.desc))
		return t

	def ss_join(
			self,
			tdir,
			ofmt,
			in_exts,
			add_args = [],
			sid      = None,
			exit_val = None,
			master   = None,
			id_str   = None):
		td = self.get_tmp_subdir(tdir)
		shares = [get_file_with_ext(td,f) for f in in_exts]
		if not sid:
			sid = self.read_from_tmpfile('dfl.sid')
		t = self.spawn('mmgen-seedjoin',
				add_args
				+ ([f'--master-share={master}'] if master else [])
				+ ([f'--id-str={id_str}'] if id_str else [])
				+ ['-d',td,'-o',ofmt]
				+ (['--label','Joined Wallet Label','-r0'] if ofmt == 'w' else [])
				+ shares,
				exit_val = exit_val)
		if exit_val:
			return t
		icls = ( dfl_wcls if 'mmdat' in in_exts
			else get_wallet_cls('incog') if 'mmincog' in in_exts
			else get_wallet_cls('incog_hex') if 'mmincox' in in_exts
			else get_wallet_cls('incog_hidden') if '-H' in add_args
			else None )
		if icls.type.startswith('incog'):
			t.hash_preset(icls.desc,'1')
		if icls:
			t.passphrase(icls.desc,sh1_passwd)
		if master:
			fs = "master share #{}, split id.*‘{}’.*, share count {}"
			pat = fs.format(
				master,
				id_str or 'default',
				len(shares) + (icls.type=='incog_hidden') )
			t.expect(pat,regex=True)
		sid_cmp = strip_ansi_escapes(t.expect_getend('Joined Seed ID: '))
		cmp_or_die(sid,sid_cmp)
		ocls = get_wallet_cls(fmt_code=ofmt)
		if ocls.type == 'mmgen':
			t.hash_preset('new '+ocls.desc,'1')
			t.passphrase_new('new '+ocls.desc,wpasswd)
		t.written_to_file(capfirst(ocls.desc))
		return t

	def get_hincog_arg(self,tdir,suf='-default-2of2'):
		sid = self.read_from_tmpfile('dfl.sid')
		return os.path.join(self.tmpdir,tdir,sid+suf+'.hincog') + ',123'

	def ss_2way_A_dfl1(self):
		return self.ss_splt('2way_dfl1','w','1:2')
	def ss_2way_B_dfl1(self):
		return self.ss_splt('2way_dfl1','bip39','2:2')
	def ss_2way_join_dfl1(self):
		return self.ss_join('2way_dfl1','w',['mmdat','bip39'])

	def ss_2way_A_dfl2(self):
		return self.ss_splt('2way_dfl2','seed','default:1:2')
	def ss_2way_B_dfl2(self):
		return self.ss_splt('2way_dfl2','hincog','default:2:2',['-J',self.get_hincog_arg('2way_dfl2')])
	def ss_2way_join_dfl2(self):
		return self.ss_join('2way_dfl2','mmhex',['mmseed'],['-H',self.get_hincog_arg('2way_dfl2')])

	def ss_2way_A_alice(self):
		return self.ss_splt('2way_alice','w','alice:1:2')
	def ss_2way_B_alice(self):
		return self.ss_splt('2way_alice','mmhex','alice:2:2')
	def ss_2way_join_alice(self):
		return self.ss_join('2way_alice','seed',['mmdat','mmhex'])
	def ss_2way_join_alice_mix(self):
		return self.ss_join('2way_alice','seed',['mmhex','mmdat'])

	def ss_2way_A_dfl_usw(self):
		return self.ss_splt('2way_dfl_usw','words','1:2',[],wf=ref_wf)
	def ss_2way_B_dfl_usw(self):
		return self.ss_splt('2way_dfl_usw','incog','2:2',[],wf=ref_wf)
	def ss_2way_join_dfl_usw(self):
		return self.ss_join('2way_dfl_usw','mmhex',['mmwords','mmincog'],sid=ref_sid)

	def ss_3way_A_dfl(self):
		return self.ss_splt('3way_dfl','words','1:3')
	def ss_3way_B_dfl(self):
		return self.ss_splt('3way_dfl','incog_hex','2:3')
	def ss_3way_C_dfl(self):
		return self.ss_splt('3way_dfl','bip39','3:3')
	def ss_3way_join_dfl(self):
		return self.ss_join('3way_dfl','mmhex',['mmwords','mmincox','bip39'])
	def ss_3way_join_dfl_mix(self):
		return self.ss_join('3way_dfl','mmhex',['bip39','mmwords','mmincox'])

	def ss_2way_A_dfl_master3(self):
		return self.ss_splt('2way_dfl_master3','w','',master=3)
	def ss_2way_B_dfl_master3(self):
		return self.ss_splt('2way_dfl_master3','bip39','2:2',master=3)
	def ss_2way_join_dfl_master3(self):
		return self.ss_join('2way_dfl_master3','mmhex',['mmdat','bip39'],master=3)

	tdir2 = '3way_foobar_master7'
	def ss_3way_C_foobar_master7(self):
		return self.ss_splt(self.tdir2,'hincog','',
					['-J',self.get_hincog_arg(self.tdir2,'-master7')],master=7)
	def ss_3way_B_foobar_master7(self):
		return self.ss_splt(self.tdir2,'bip39','φυβαρ:2:3',master=7)
	def ss_3way_A_foobar_master7(self):
		return self.ss_splt(self.tdir2,'mmhex','φυβαρ:3:3',master=7)
	def ss_3way_join_foobar_master7(self):
		return self.ss_join(self.tdir2,'seed', ['bip39','mmhex'],
							['-H',self.get_hincog_arg(self.tdir2,'-master7')],master=7,id_str='φυβαρ')
	def ss_3way_join_foobar_master7_mix(self):
		return self.ss_join(self.tdir2,'seed', ['mmhex','bip39'],
							['-H',self.get_hincog_arg(self.tdir2,'-master7')],master=7,id_str='φυβαρ')

	def ss_bad_invocation(self,cmd,args,exit_val,errmsg):
		t = self.spawn(cmd, args, exit_val=exit_val)
		t.expect(errmsg, regex=True)
		return t

	def ss_3way_join_dfl_bad_invocation(self):
		t = self.ss_join('3way_dfl','mmhex',
			['mmwords','mmincox','bip39'],
			id_str   = 'foo',
			exit_val = 1)
		t.expect('option meaningless')
		return t

	def ss_bad_invocation1(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', [], 1, 'USAGE:')

	def ss_bad_invocation2(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', ['-M1', '1:9'], 1, 'meaningless in master share context')

	def ss_bad_invocation3(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', [self.tmpdir+'/no.mmdat', '1:9'], 1, 'input file .* not found')

	def ss_bad_invocation4(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', [self.tmpdir+'/dfl.sid', '1:9'], 1, 'unrecognized .* extension')

	def ss_bad_invocation5(self):
		return self.ss_bad_invocation(
			'mmgen-seedjoin', [], 1, 'USAGE:')

	def ss_bad_invocation6(self):
		return self.ss_bad_invocation(
			'mmgen-seedjoin', [self.tmpdir+'/a'], 1, 'USAGE:')

	def ss_bad_invocation7(self):
		return self.ss_bad_invocation(
			'mmgen-seedjoin', [self.tmpdir+'/a', self.tmpdir+'/b'], 1, 'unrecognized .* extension')

	def ss_bad_invocation8(self):
		return self.ss_bad_invocation(
			'mmgen-seedjoin', [self.tmpdir+'/a.mmdat', self.tmpdir+'/b.mmdat'], 1, 'input file .* not found')

	def ss_bad_invocation9(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', ['x'], 1, 'USAGE:')

	def ss_bad_invocation10(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', [self.tmpdir+'/a.mmdat', '1:2'], 1, 'input file .* not found')

	def ss_bad_invocation11(self):
		return self.ss_bad_invocation(
			'mmgen-seedsplit', [self.tmpdir+'/dfl.sid', '1:2'], 1, 'unrecognized .* extension')



#!/usr/bin/env python3

from mmgen.cfg import Config
from mmgen.util import msg

opts_data = {
	'sets': [('print_checksum',True,'quiet',True)],
	'text': {
		'desc': 'Opts test',
		'usage':'[args] [opts]',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long options (common options)
-i, --in-fmt=      f  Input is from wallet format 'f'
-d, --outdir=      d  Use outdir 'd'
-C, --print-checksum  Print a checksum
-E, --fee-estimate-mode=M Specify the network fee estimate mode.
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-k, --keep-label      Reuse label of input wallet for output wallet
-l, --seed-len=    l  Specify wallet seed length of 'l' bits.
-L, --label=       l  Specify a label 'l' for output wallet
-m, --minconf=     n  Minimum number of confirmations required to spend
                      outputs (default: 1)
-p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
-P, --passwd-file= f  Get wallet passphrase from file 'f'
-q, --quiet           Be quieter
-X, --cached-balances Use cached balances (Ethereum only)
-v, --verbose         Be more verbose
                      sample help_note: {kgs}
                      sample help_note: {coin_id}
""",
	'notes': """

                           NOTES FOR THIS COMMAND

sample note: {nn}
"""
	},
	'code': {
		'options': lambda cfg,help_notes,s: s.format(
			kgs=help_notes('keygen_backends'),
			coin_id=help_notes('coin_id'),
		),
		'notes': lambda s: s.format(nn='a note'),
	}
}

cfg = Config(opts_data=opts_data)

for k in (
	'foo',               # added opt
	'print_checksum',    # sets 'quiet'
	'quiet','verbose',   # _incompatible_opts
	'passwd_file',       # _infile_opts - check_infile()
	'outdir',            # check_outdir()
	'cached_balances',   # opt_sets_global
	'minconf',           # global_sets_opt
	'hidden_incog_input_params',
	):
	msg('{:30} {}'.format( f'cfg.{k}:', getattr(cfg,k) ))

msg('')
for k in (
	'cached_balances',   # opt_sets_global
	'minconf',           # global_sets_opt
	):
	msg('{:30} {}'.format( f'cfg.{k}:', getattr(cfg,k) ))

msg('')
for k in (
	'fee_estimate_mode', # _autoset_opts
	):
	msg('{:30} {}'.format( f'cfg.{k}:', getattr(cfg,k) ))



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_py_d.ot_btc_mainnet: BTC mainnet test vectors for MMGen data objects
"""

from decimal import Decimal

from mmgen.obj import Int,MMGenTxID,CoinTxID,MMGenWalletLabel,MMGenTxComment,MMGenPWIDString,TrackingWalletName
from mmgen.addrlist import AddrIdx,AddrIdxList,AddrListID
from mmgen.seed import Seed,SeedID
from mmgen.subseed import SubSeedList,SubSeedIdx,SubSeedIdxRange
from mmgen.seedsplit import SeedShareIdx,SeedShareCount,MasterShareIdx,SeedSplitSpecifier
from mmgen.key import PrivKey,WifKey,PubKey
from mmgen.amt import BTCAmt
from mmgen.addr import CoinAddr,MMGenID,MMGenAddrType,MMGenPasswordType
from mmgen.tw.shared import TwMMGenID,TwLabel,TwComment
from mmgen.rpc import IPPort
from mmgen.protocol import init_proto

from .ot_common import r16,r32
from ..include.common import cfg,utf8_ctrl,utf8_text,utf8_combining,gr_uc_w_ctrl,text_jp,text_zh,ru_uc,gr_uc

proto = init_proto( cfg, 'btc', need_amt=True )
tw_pfx = proto.base_coin.lower() + ':'
zero_addr = '1111111111111111111114oLvT2'

ssm = str(SeedShareCount.max_val)
privkey = PrivKey(proto=proto,s=bytes.fromhex('deadbeef'*8),compressed=True,pubkey_type='std')

tests = {
	'Int': {
		'arg1': 'n',
		'bad':  ('1L',0.0,'0.0','1.0',1.0,'s',1.1,'1.1'),
		'good': (
			('0',0),('-1',-1),('7',7),-1,0,1,9999999,
			{'n':'0x0','base':16,'ret':0},
			{'n':'0x1','base':16,'ret':1},
			{'n':'0xf','base':16,'ret':15},
			{'n':'0xff','base':16,'ret':255},
		)
	},
	'AddrIdx': {
		'arg1': 'n',
		'bad':  ('s',1.1,10000000,-1,0),
		'good': (('7',7),(1,1),(9999999,9999999))
	},
	'SeedShareIdx': {
		'arg1': 'n',
		'bad':  ('s',1.1,1025,-1,0),
		'good': (('7',7),(1,1),(1024,1024))
	},
	'SeedShareCount': {
		'arg1': 'n',
		'bad':  ('s',2.1,1025,-1,0,1),
		'good': (('7',7),(2,2),(1024,1024))
	},
	'MasterShareIdx': {
		'arg1': 'n',
		'bad':  ('s',1.1,1025,-1,0),
		'good': (('7',7),(1,1),(1024,1024))
	},
	'AddrIdxList': {
		'arg1': 'fmt_str',
		'bad':  ('x','5,9,1-2-3','8,-11','66,3-2','0-3'),
		'good': (
			('3,2,2',(2,3)),
			('101,1,3,5,2-7,99',(1,2,3,4,5,6,7,99,101)),
			({'idx_list':AddrIdxList('1-5')},(1,2,3,4,5))
		)
	},
	'SubSeedIdxRange': {
		'bad':  (33,'x','-11','66,3','0','3-2','8000000','100000000',(1,2,3)),
		'good': (
			('3',(3,3)),
			((3,5),(3,5)),
			('1-2',(1,2)),
			(str(SubSeedList.dfl_len),(SubSeedList.dfl_len,SubSeedList.dfl_len)),
			(str(SubSeedIdxRange.max_idx),(SubSeedIdxRange.max_idx,SubSeedIdxRange.max_idx)),
		)
	},
	'BTCAmt': {
		'arg1': 'num',
		'bad':  ('-3.2','0.123456789',123,'123L','22000000',20999999.12345678,
					{'num':'1','from_decimal':True},
					{'num':1,'from_decimal':True},
				),
		'good': (('20999999.12345678',Decimal('20999999.12345678')),
				{'num':Decimal('1.23456789623456789'),'from_decimal':True,'ret':Decimal('1.23456790')}, # rounding
				{'num':Decimal('1.234'),'from_decimal':True,'ret':Decimal('1.234')},
				{'num':Decimal('0.0'),'from_decimal':True,'ret':Decimal('0')},
				# emulate network fee estimation:
				#                  BTC/kB         fee_adjust       tx size
				{   'num':Decimal('0.00053249') * Decimal('0.9') * 109 / 1024 , # ≈53 sat/byte
					'from_decimal':True,
					'ret':Decimal('0.00005101') },
				{   'num':Decimal('0.00003249') * Decimal('1.1') * 109 / 1024 , # ≈3 sat/byte
					'from_decimal':True,
					'ret':Decimal('0.00000380') },
				{   'num':Decimal('0.00011249') * Decimal('1.0') * 221 / 1024 , # ≈11 sat/byte
					'from_decimal':True,
					'ret':Decimal('0.00002428') },
				{'num':1234,'from_unit':'satoshi','ret':Decimal('0.00001234')},
		)
	},
	'CoinAddr': {
		'arg1': 'addr',
		'good':  (
			{'addr':'1MjjELEy6EJwk8fSNfpS8b5teFRo4X5fZr', 'proto':proto},
			{'addr':'32GiSWo9zJQgkCmjAaLRrbPwXhKry2jHhj', 'proto':proto},
		),
		'bad':  (
			{'addr':1,   'proto':proto},
			{'addr':'x', 'proto':proto},
			{'addr':'я', 'proto':proto},
		),
	},
	'SeedID': {
		'arg1': 'sid',
		'bad':  (
			{'sid':'я'},
			{'sid':'F00F00'},
			{'sid':'xF00F00x'},
			{'sid':1},
			{'sid':'F00BAA123'},
			{'sid':'f00baa12'},
			'я',r32,'abc'
			),
		'good': (
			{'sid':'F00BAA12'},
			{'seed': Seed(cfg,r16),     'ret': SeedID(seed=Seed(cfg,r16))},
			{'sid':  Seed(cfg,r16).sid, 'ret': SeedID(seed=Seed(cfg,r16))}
			)
	},
	'SubSeedIdx': {
		'arg1': 's',
		'bad':  (33,'x','я','1x',200,'1ss','L','s','200LS','30ll','s100',str(SubSeedIdxRange.max_idx+1),'0'),
		'good': (('1','1L'),('1s','1S'),'20S','30L',('300l','300L'),('200','200L'),str(SubSeedIdxRange.max_idx)+'S')
	},
	'MMGenID': {
		'arg1': 'id_str',
		'bad':  (
			{'id_str':'x',             'proto':proto},
			{'id_str':1,               'proto':proto},
			{'id_str':'f00f00f',       'proto':proto},
			{'id_str':'a:b',           'proto':proto},
			{'id_str':'x:L:3',         'proto':proto},
			{'id_str':'F00BAA12',      'proto':proto},
			{'id_str':'F00BAA12:Z:99', 'proto':proto},
		),
		'good':  (
			{'id_str':'F00BAA12:99',   'proto':proto, 'ret':'F00BAA12:L:99'},
			{'id_str':'F00BAA12:L:99', 'proto':proto},
			{'id_str':'F00BAA12:S:99', 'proto':proto},
		),
	},
	'TwMMGenID': {
		'arg1': 'id_str',
		'bad':  (
			{'id_str':'x',             'proto':proto},
			{'id_str':'я',             'proto':proto},
			{'id_str':'я:я',           'proto':proto},
			{'id_str':1,               'proto':proto},
			{'id_str':'f00f00f',       'proto':proto},
			{'id_str':'a:b',           'proto':proto},
			{'id_str':'x:L:3',         'proto':proto},
			{'id_str':'F00BAA12:0',    'proto':proto},
			{'id_str':'F00BAA12:Z:99', 'proto':proto},
			{'id_str':tw_pfx,          'proto':proto},
			{'id_str':tw_pfx+'я',      'proto':proto},
			{'id_str':tw_pfx+'x',      'proto':proto},
		),
		'good':  (
			{'id_str':tw_pfx+zero_addr,     'proto':proto},
			{'id_str':'F00BAA12:99',        'proto':proto, 'ret':'F00BAA12:L:99'},
			{'id_str':'F00BAA12:L:99',      'proto':proto},
			{'id_str':'F00BAA12:S:9999999', 'proto':proto},
		),
	},
	'TrackingWalletName': {
		'arg1': 's',
		'bad':  ('-abcdefg', 'a' * 50, 'abc-α-ω'),
		'good': ('ab-cd-def_ghi-', '321-abcd')
	},
	'TwLabel': {
		'arg1': 'proto',
		'exc_name': 'BadTwLabel',
		'bad':  (
			{'text':'',              'proto':proto},
			{'text':'x x',           'proto':proto},
			{'text':'x я',           'proto':proto},
			{'text':'я:я',           'proto':proto},
			{'text':1,               'proto':proto},
			{'text':'f00f00f',       'proto':proto},
			{'text':'a:b',           'proto':proto},
			{'text':'x:L:3',         'proto':proto},
			{'text':'F00BAA12:0 x',  'proto':proto},
			{'text':'F00BAA12:Z:99', 'proto':proto},
			{'text':tw_pfx+' x',     'proto':proto},
			{'text':tw_pfx+'я x',    'proto':proto},
			{'text':utf8_ctrl[:40],  'proto':proto},
			{'text':'F00BAA12:S:1 ' + utf8_ctrl[:40], 'proto':proto, 'exc_name': 'BadTwComment'},
			{'text':tw_pfx+'x comment','proto':proto},
		),
		'good':  (
			{'text':'F00BAA12:99 a comment',            'proto':proto, 'ret':'F00BAA12:L:99 a comment'},
			{'text':'F00BAA12:L:99 a comment',          'proto':proto},
			{'text': 'F00BAA12:L:99 comment (UTF-8) α', 'proto':proto},
			{'text':'F00BAA12:S:9999999 comment',       'proto':proto},
			{'text':tw_pfx+zero_addr+' comment',        'proto':proto},
		),
	},
	'MMGenTxID': {
		'arg1': 's',
		'bad':  (1,[],'\0','\1','я','g','gg','FF','f00','F00F0012'),
		'good': ('DEADBE','F00BAA')
	},
	'CoinTxID':{
		'arg1': 's',
		'bad':  (1,[],'\0','\1','я','g','gg','FF','f00','F00F0012',r16.hex(),r32.hex()+'ee'),
		'good': (r32.hex(),)
	},
	'WifKey': {
		'arg1': 'proto',
		'bad': (
			{'proto':proto, 'wif':1},
			{'proto':proto, 'wif':[]},
			{'proto':proto, 'wif':'\0'},
			{'proto':proto, 'wif':'\1'},
			{'proto':proto, 'wif':'я'},
			{'proto':proto, 'wif':'g'},
			{'proto':proto, 'wif':'gg'},
			{'proto':proto, 'wif':'FF'},
			{'proto':proto, 'wif':'f00'},
			{'proto':proto, 'wif':r16.hex()},
			{'proto':proto, 'wif':'2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN'},
		),
		'good': (
			{'proto':proto, 'wif':'5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb',  'ret_idx':1},
			{'proto':proto, 'wif':'KwWr9rDh8KK5TtDa3HLChEvQXNYcUXpwhRFUPc5uSNnMtqNKLFhk', 'ret_idx':1},
		)
	},
	'PubKey': {
		'arg1': 's',
		'bad':  (
			{'s':1, 'compressed':True },
			{'s':'F00BAA12','compressed':False},
		),
		'good': ( # TODO: add real pubkeys
			{'s':bytes.fromhex('deadbeef'),'compressed':True},
		)
	},
	'PrivKey': {
		'arg1': 'proto',
		'bad': (
			{'proto':proto, 'wif':1},
			{'proto':proto, 'wif':'1'},
			{'proto':proto, 'wif':'cMsqcmDYZP1LdKgqRh9L4ZRU9br28yvdmTPwW2YQwVSN9aQiMAoR'},
			{'proto':proto, 's':r32,'wif':'5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb'},
			{'proto':proto, 'pubkey_type':'std'},
			{'proto':proto, 's':r32},
			{'proto':proto, 's':r32,'compressed':'yes'},
			{'proto':proto, 's':r32,'compressed':'yes','pubkey_type':'std'},
			{'proto':proto, 's':r32,'compressed':True,'pubkey_type':'nonstd'},
			{'proto':proto, 's':r32+b'x','compressed':True,'pubkey_type':'std'}
		),
		'good': (
			{'proto':proto, 'wif':'5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb',
			'ret':bytes.fromhex('e0aef965b905a2fedf907151df8e0a6bac832aa697801c51f58bd2ecb4fd381c')},
			{'proto':proto, 'wif':'KwWr9rDh8KK5TtDa3HLChEvQXNYcUXpwhRFUPc5uSNnMtqNKLFhk',
			'ret':bytes.fromhex('08d0ed83b64b68d56fa064be48e2385060ed205be2b1e63cd56d218038c3a05f')},
			{'proto':proto, 's':r32,'compressed':False,'pubkey_type':'std','ret':r32},
			{'proto':proto, 's':r32,'compressed':True,'pubkey_type':'std','ret':r32}
		)
	},
	'AddrListID': {
		'arg1': 'sid',
		'bad':  (
			{'sid':SeedID(sid='F00BAA12'),'mmtype':'Z','ret':'F00BAA12:Z'},
			{'id_str':'G00BAA12:B','proto':proto},
		),
		'good':  (
			{'id_str':'F00BAA12:B','proto':proto,'ret':'F00BAA12:B'},
			{'sid':SeedID(sid='F00BAA12'),'mmtype':proto.addr_type(id_str='S'),'ret':'F00BAA12:S'},
			{'sid':SeedID(sid='F00BAA12'),'mmtype':proto.addr_type(id_str='L'),'ret':'F00BAA12:L'},
		)
	},
	'MMGenWalletLabel': {
		'arg1': 's',
		'bad': (utf8_text[:49],utf8_combining[:48],utf8_ctrl[:48],gr_uc_w_ctrl),
		'good':  (utf8_text[:48],)
	},
	'TwComment': {
		'exc_name': 'BadTwComment',
		'arg1': 's',
		'bad': (    utf8_combining[:40],
					utf8_ctrl[:40],
					text_jp[:41],
					text_zh[:41],
					gr_uc_w_ctrl,
					utf8_text[:81] ),
		'good': (   utf8_text[:80],
					(ru_uc + gr_uc + utf8_text)[:80],
					text_jp[:40],
					text_zh[:40] )
	},
	'MMGenTxComment':{
		'arg1': 's',
		'bad': (utf8_text[:73],utf8_combining[:72],utf8_ctrl[:72],gr_uc_w_ctrl),
		'good':  (utf8_text[:72],)
	},
	'MMGenPWIDString': { # forbidden = list(u' :/\\')
		'arg1': 's',
		'bad': ('foo/','foo:','foo:\\'),
		'good': ('qwerty@яяя',)
	},
	'MMGenAddrType': {
		'arg1': 'proto',
		'bad':  (
			{'proto':proto, 'id_str':'U',        'ret':'L'},
			{'proto':proto, 'id_str':'z',        'ret':'L'},
			{'proto':proto, 'id_str':'xx',       'ret':'C'},
			{'proto':proto, 'id_str':'dogecoin', 'ret':'C'},
		),
		'good':  (
			{'proto':proto, 'id_str':'legacy',    'ret':'L'},
			{'proto':proto, 'id_str':'L',         'ret':'L'},
			{'proto':proto, 'id_str':'compressed','ret':'C'},
			{'proto':proto, 'id_str':'C',         'ret':'C'},
			{'proto':proto, 'id_str':'segwit',    'ret':'S'},
			{'proto':proto, 'id_str':'S',         'ret':'S'},
			{'proto':proto, 'id_str':'bech32',    'ret':'B'},
			{'proto':proto, 'id_str':'B',         'ret':'B'}
		)
	},
	'MMGenPasswordType': {
		'arg1': 'proto',
		'bad':  (
			{'proto':proto, 'id_str':'U',        'ret':'L'},
			{'proto':proto, 'id_str':'z',        'ret':'L'},
			{'proto':proto, 'id_str':'я',        'ret':'C'},
			{'proto':proto, 'id_str':1,          'ret':'C'},
			{'proto':proto, 'id_str':'passw0rd', 'ret':'C'},
		),
		'good': (
			{'proto':proto, 'id_str':'password', 'ret':'P'},
			{'proto':proto, 'id_str':'P',        'ret':'P'},
		)
	},
	'SeedSplitSpecifier': {
		'arg1': 's',
		'bad': ('M','αβ:2',1,'0:1','1:1','2:1','3:2','1:2000','abc:0:2'),
		'good': (
			('1:2','2:2','alice:2:2','αβ:2:2','1:'+ssm,ssm+':'+ssm)
		)
	},
	'IPPort': {
		'arg1': 's',
		'bad': (
			'.0.0.0:0',
			'127.0.0.1',
			'127.0.0.1:',
			'20',
			':20',
			'127.0.0.256:0',
			'127.0.0.1:65536',
			'127.0.0.01:60000',
			'0.0.0.0:00',
		),
		'good': (
			(
				'0.0.0.0:0',
				'0.0.0.1:0',
				'0.0.1.0:0',
				'0.0.1.1:0',
				'0.1.0.0:0',
				'1.0.0.0:0',
				'10.0.0.1:0',
				'127.0.0.1:65535',
				'255.0.0.0:400',
				'255.255.255.255:65535',
			)
		)
	},
}



#!/usr/bin/env python3

"""
test/unit_tests_d/ut_seedsplit: seed splitting unit test for the MMGen suite
"""

from mmgen.util import msg,msg_r

from ..include.common import cfg,vmsg,vmsg_r

class unit_test:

	def run_test(self,name,ut):
		from mmgen.seed import Seed
		from mmgen.seedsplit import SeedShareList,SeedShareIdx

		cfg.debug_subseed = bool(cfg.verbose)

		def basic_ops(master_idx):
			test_data = {
				'default': (
					(8,'4710FBF0','B3D9411B','2670E83D','D1FC57ED','AE49CABE','63FFBA62',0,0),
					(6,'9D07ABBD','AF5DC2F6','1A3BBDAC','2548AEE9','B94F7450','1F4E5A12',0,0),
					(4,'43670520','1F72C066','E5AA8DA1','A33966A0','D2BCE0A5','A568C315',0,0),
				),
				'φυβαρ': (
					(8,'4710FBF0','269D658C','9D25889E','6D730ECB','C61A963F','9FE99C05',0,0),
					(6,'9D07ABBD','4998B33E','F00CE041','C612BEE5','35CD3675','41B3BE61',0,0),
					(4,'43670520','77140076','EA82CB30','80F7AEDE','D168D768','77BE57AA',0,0),
				)
			}
			test_data_master = {
				'1': {
					'default': (
						(8,'4710FBF0','6AE6177F','AC12090C','6AE6177F','3E87A907','7D1FEA56','BFEBFFFF','629A9808'),
						(4,'43670520','6739535C','ABF4DD38','6739535C','778E9C60','89CBCFD2','689FABF5','70BED76B'),
					),
					'φυβαρ': (
						(8,'4710FBF0','6AE6177F','AC5FA32E','6AE6177F','9777A750','C7CF2AFC','035AAACB','C777FBE4'),
						(4,'43670520','6739535C','37EBA2F5','6739535C','927549D2','29BADEE7','9CA73A03','313F5528'))
				},
				'5': {
					'default': (
						(8,'4710FBF0','5EFAC3D6','B489167D','5EFAC3D6','BB004DC5','1A0381C0','4EA182E3','547FB2DC'),
						(4,'43670520','EE93DB0E','44962A7D','EE93DB0E','07339882','376A05B1','CE51D022','00149CA3'),
					),
					'φυβαρ': (
						(8,'4710FBF0','5EFAC3D6','A6E27EE3','5EFAC3D6','32C24668','B4C54297','1EC9B71B','8C5C6B1C'),
						(4,'43670520','EE93DB0E','B584E963','EE93DB0E','4BEA2AB2','4BEA65C7','140FC43F','BBD19461'))
				}
			}
			if master_idx:
				test_data = test_data_master[str(master_idx)]

			for id_str in (None,'default','φυβαρ'):
				msg_r(f'Testing basic ops (id_str={id_str!r}, master_idx={master_idx})...')
				vmsg('')

				for a,b,c,d,e,f,h,i,p in test_data[id_str if id_str is not None else 'default']:
					seed_bin = bytes.fromhex('deadbeef' * a)
					seed = Seed( cfg, seed_bin )
					assert seed.sid == b, seed.sid

					for share_count,j,k,l,m in ((2,c,c,d,i),(5,e,f,h,p)):

						shares = seed.split(share_count,id_str,master_idx)
						A = len(shares)
						assert A == share_count, A

						s = shares.format()
						vmsg_r(f'\n{s}')
						assert len(s.strip().split('\n')) == share_count+6, s

						if master_idx:
							A = shares.get_share_by_idx(1,base_seed=False).sid
							B = shares.get_share_by_seed_id(j,base_seed=False).sid
							assert A == B == m, A

						A = shares.get_share_by_idx(1,base_seed=True).sid
						B = shares.get_share_by_seed_id(j,base_seed=True).sid
						assert A == B == j, A

						A = shares.get_share_by_idx(share_count-1,base_seed=True).sid
						B = shares.get_share_by_seed_id(k,base_seed=True).sid
						assert A == B == k, A

						A = shares.get_share_by_idx(share_count).sid
						B = shares.get_share_by_seed_id(l).sid
						assert A == B == l, A

						A = shares.join().sid
						assert A == b, A

						if master_idx:
							slist = [shares.get_share_by_idx(i+1,base_seed=True) for i in range(len(shares))]
							A = Seed.join_shares( cfg, slist, master_idx, id_str ).sid
							assert A == b, A

				msg('OK')

		def defaults_and_limits():
			msg_r('Testing defaults and limits...')

			seed_bin = bytes.fromhex('deadbeef' * 8)
			seed = Seed( cfg, seed_bin )

			shares = seed.split(SeedShareIdx.max_val)
			s = shares.format()
#			vmsg_r(f'\n{s}')
			assert len(s.strip().split('\n')) == 1030, s

			A = shares.get_share_by_idx(1024).sid
			B = shares.get_share_by_seed_id('4BA23728').sid
			assert A == '4BA23728', A
			assert B == '4BA23728', B

			A = shares.join().sid
			B = seed.sid
			assert A == B, A

			msg('OK')

		def collisions(seed_hex,ss_count,last_sid,collisions_chk,master_idx):

			msg_r(f'Testing Seed ID collisions ({ss_count} seed shares, master_idx={master_idx})...')
			vmsg('')

			seed_bin = bytes.fromhex(seed_hex)
			seed = Seed( cfg, seed_bin )

			SeedShareIdx.max_val = ss_count
			shares = seed.split(ss_count,master_idx=master_idx)
			A = shares.get_share_by_idx(ss_count).sid
			B = shares.get_share_by_seed_id(last_sid).sid
			assert A == last_sid, A
			assert B == last_sid, B

			assert shares.nonce_start == 0, shares.nonce_start

			collisions = 0
			for sid in shares.data['long']:
				collisions += shares.data['long'][sid][1]

			assert collisions == collisions_chk, collisions
			vmsg_r(f'{collisions} collisions, last_sid {last_sid}')
			msg('OK')

		def last_share_collisions():
			msg_r('Testing last share collisions with shortened Seed IDs')
			vmsg('')
			seed_bin = bytes.fromhex('2eadbeef'*8)
			seed = Seed( cfg, seed_bin )
			ssm_save = SeedShareIdx.max_val
			ssm = SeedShareIdx.max_val = 2048
			shares = SeedShareList(seed,count=ssm,id_str='foo',master_idx=1,debug_last_share=True)
			lsid = shares.last_share.sid
			collisions = shares.data['long'][lsid][1]
			assert collisions == 2, collisions
			assert lsid == 'B5B8AD09', lsid
			SeedShareIdx.max_val = ssm_save
			vmsg_r(f'{collisions} collisions, last_share sid {lsid}')
			msg('..OK')

		basic_ops(master_idx=None)
		basic_ops(master_idx=1)
		basic_ops(master_idx=5)
		defaults_and_limits()
		last_share_collisions()
		collisions('1dabcdef'*4,65535,'B5CBCE0A',3,master_idx=None)
		collisions('18abcdef'*4,65535,'FF03CE82',3,master_idx=1)

		return True



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_py_d.ct_opts: options processing tests for the MMGen cmdtest.py test suite
"""

import os,time

from ..include.common import cfg
from .ct_base import CmdTestBase

class CmdTestOpts(CmdTestBase):
	'options processing'
	networks = ('btc',)
	tmpdir_nums = [41]
	cmd_group = (
		('opt_helpscreen',       (41,"helpscreen output", [])),
		('opt_noargs',           (41,"invocation with no user options or arguments", [])),
		('opt_good',             (41,"good opts", [])),
		('opt_bad_infile',       (41,"bad infile parameter", [])),
		('opt_bad_outdir',       (41,"bad outdir parameter", [])),
		('opt_bad_incompatible', (41,"incompatible opts", [])),
		('opt_bad_autoset',      (41,"invalid autoset value", [])),
	)

	def spawn_prog(self, args, exit_val=None):
		return self.spawn('test/misc/opts.py', args, cmd_dir='.', exit_val=exit_val)

	def check_vals(self,args,vals):
		t = self.spawn_prog(args)
		for k,v in vals:
			t.expect(rf'{k}:\s+{v}',regex=True)
		return t

	def do_run(self,args,expect,exit_val,regex=False):
		t = self.spawn_prog(args, exit_val=exit_val or None)
		t.expect(expect,regex=regex)
		return t

	def opt_helpscreen(self):
		expect = r'OPTS.PY: Opts test.*USAGE:\s+opts.py'
		if not cfg.pexpect_spawn:
			expect += r'.*--minconf.*NOTES FOR THIS.*a note'
		t = self.do_run( ['--help'], expect, 0, regex=True )
		if t.pexpect_spawn:
			time.sleep(0.4)
			t.send('q')
		return t

	def opt_noargs(self):
		return self.check_vals(
				[],
				(
					('cfg.foo',                 'None'),         # added opt
					('cfg.print_checksum',      'None'),         # sets 'quiet'
					('cfg.quiet',               'False'),        # _incompatible_opts
					('cfg.verbose',             'False'),        # _incompatible_opts
					('cfg.passwd_file',         ''),             # _infile_opts - check_infile()
					('cfg.outdir',              ''),             # check_outdir()
					('cfg.cached_balances',     'False'),
					('cfg.minconf',             '1'),
					('cfg.fee_estimate_mode',   'conservative'), # _autoset_opts
				)
			)

	def opt_good(self):
		pf_base = 'testfile'
		pf = os.path.join(self.tmpdir,pf_base)
		self.write_to_tmpfile(pf_base,'')
		return self.check_vals(
				[
					'--print-checksum',
					'--fee-estimate-mode=E',
					'--passwd-file='+pf,
					'--outdir='+self.tmpdir,
					'--cached-balances',
					f'--hidden-incog-input-params={pf},123',
				],
				(
					('cfg.print_checksum',           'True'),
					('cfg.quiet',                    'True'), # set by print_checksum
					('cfg.passwd_file',              pf),
					('cfg.outdir',                   self.tmpdir),
					('cfg.cached_balances',          'True'),
					('cfg.hidden_incog_input_params', pf+',123'),
					('cfg.fee_estimate_mode',         'economical'),
				)
			)

	def opt_bad_infile(self):
		pf = os.path.join(self.tmpdir,'fubar')
		return self.do_run(['--passwd-file='+pf],'not found',1)

	def opt_bad_outdir(self):
		bo = self.tmpdir+'_fubar'
		return self.do_run(['--outdir='+bo],'not found',1)

	def opt_bad_incompatible(self):
		return self.do_run(['--label=Label','--keep-label'],'Conflicting options',1)

	def opt_bad_autoset(self):
		return self.do_run(['--fee-estimate-mode=Fubar'],'not unique substring',1)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.params: Ethereum protocol
"""

from ...protocol import CoinProtocol,_nw,decoded_addr
from ...util import is_hex_str_lc,Msg

class mainnet(CoinProtocol.DummyWIF,CoinProtocol.Secp256k1):

	network_names = _nw('mainnet','testnet','devnet')
	addr_len      = 20
	mmtypes       = ('E',)
	dfl_mmtype    = 'E'
	mod_clsname   = 'Ethereum'
	pubkey_type   = 'std' # required by DummyWIF

	coin_amt      = 'ETHAmt'
	max_tx_fee    = '0.005'
	chain_names   = ['ethereum','foundation']
	sign_mode     = 'standalone'
	caps          = ('token',)
	mmcaps        = ('key','addr','rpc_init','tx')
	base_proto    = 'Ethereum'
	base_proto_coin = 'ETH'
	base_coin     = 'ETH'
	avg_bdi       = 15
	ignore_daemon_version = False

	chain_ids = {
		1:    'ethereum',         # ethereum mainnet
		2:    'morden',           # morden testnet (deprecated)
		3:    'ropsten',          # ropsten testnet
		4:    'rinkeby',          # rinkeby testnet
		5:    'goerli',           # goerli testnet
		42:   'kovan',            # kovan testnet
		61:   'classic',          # ethereum classic mainnet
		62:   'morden',           # ethereum classic testnet
		17:   'developmentchain', # parity dev chain
		1337: 'developmentchain', # geth dev chain
		711:  'ethereum',         # geth mainnet (empty chain)
	}

	@property
	def dcoin(self):
		return self.tokensym or self.coin

	def decode_addr(self,addr):
		if is_hex_str_lc(addr) and len(addr) == self.addr_len * 2:
			return decoded_addr( bytes.fromhex(addr), None, 'ethereum' )
		if self.cfg.debug:
			Msg(f'Invalid address: {addr}')
		return False

	def checksummed_addr(self,addr):
		h = self.keccak_256(addr.encode()).digest().hex()
		return ''.join(addr[i].upper() if int(h[i],16) > 7 else addr[i] for i in range(len(addr)))

	def pubhash2addr(self,pubhash,p2sh):
		assert len(pubhash) == 20, f'{len(pubhash)}: invalid length for {self.name} pubkey hash'
		assert not p2sh, f'{self.name} protocol has no P2SH address format'
		return pubhash.hex()

class testnet(mainnet):
	chain_names = ['kovan','goerli','rinkeby']

class regtest(testnet):
	chain_names = ['developmentchain']



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.unsigned: unsigned transaction class
"""

from .completed import Completed
from ..util import remove_dups

class Unsigned(Completed):
	desc = 'unsigned transaction'
	ext  = 'rawtx'
	automount = False

	def delete_attrs(self,desc,attr):
		for e in getattr(self,desc):
			if hasattr(e,attr):
				delattr(e,attr)

	def get_sids(self,desc):
		return remove_dups(
			(e.mmid.sid for e in getattr(self,desc) if e.mmid),
			quiet = True )

class AutomountUnsigned(Unsigned):
	desc = 'unsigned automount transaction'
	ext  = 'arawtx'
	automount = True



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xmr.params: Monero protocol
"""

from collections import namedtuple

from ...protocol import CoinProtocol,_nw
from ...obj import HexStr

parsed_addr = namedtuple('parsed_addr',['ver_bytes','data','payment_id'])

class MoneroViewKey(HexStr):
	color,width,hexcase = 'cyan',64,'lower' # FIXME - no checking performed

# https://github.com/monero-project/monero/blob/master/src/cryptonote_config.h
class mainnet(CoinProtocol.DummyWIF,CoinProtocol.Base):

	network_names  = _nw('mainnet','stagenet',None)
	base_proto     = 'Monero'
	base_proto_coin = 'XMR'
	base_coin      = 'XMR'
	addr_ver_info  = { '12': 'monero', '2a': 'monero_sub', '13': 'monero_integrated' }
	pubkey_types   = ('monero',)
	mmtypes        = ('M',)
	dfl_mmtype     = 'M'
	pubkey_type    = 'monero' # required by DummyWIF
	avg_bdi        = 120
	privkey_len    = 32
	mmcaps         = ('key','addr')
	ignore_daemon_version = False
	coin_amt       = 'XMRAmt'
	sign_mode      = 'standalone'

	def get_addr_len(self,addr_fmt):
		return (64,72)[addr_fmt == 'monero_integrated']

	def preprocess_key(self,sec,pubkey_type): # reduce key
		from ...contrib.ed25519 import l
		return int.to_bytes(
			int.from_bytes( sec[::-1], 'big' ) % l,
			self.privkey_len,
			'big' )[::-1]

	def decode_addr(self,addr):

		from ...baseconv import baseconv

		def b58dec(addr_str):
			bc = baseconv('b58')
			l = len(addr_str)
			a = b''.join([bc.tobytes( addr_str[i*11:i*11+11], pad=8 ) for i in range(l//11)])
			b = bc.tobytes( addr_str[-(l%11):], pad=5 )
			return a + b

		ret = b58dec(addr)

		chk = self.keccak_256(ret[:-4]).digest()[:4]

		assert ret[-4:] == chk, f'{ret[-4:].hex()}: incorrect checksum.  Correct value: {chk.hex()}'

		return self.decode_addr_bytes(ret[:-4])

	def parse_addr(self,ver_bytes,addr_bytes,fmt):
		addr_len = self.get_addr_len('monero')
		return parsed_addr(
			ver_bytes  = ver_bytes,
			data       = addr_bytes[:addr_len],
			payment_id = addr_bytes[addr_len:] if fmt == 'monero_integrated' else None,
		)

	def pubhash2addr(self,*args,**kwargs):
		raise NotImplementedError('Monero addresses do not support pubhash2addr()')

	def viewkey(self,viewkey_str):
		return MoneroViewKey.__new__(MoneroViewKey,viewkey_str)

class testnet(mainnet): # use stagenet for testnet
	# testnet is {'35','3f','36'}
	addr_ver_info = { '18': 'monero', '24': 'monero_sub', '19': 'monero_integrated' }



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.bch.params: Bitcoin Cash protocol
"""

from ..btc.params import mainnet,_finfo

class mainnet(mainnet):
	is_fork_of      = 'Bitcoin'
	mmtypes         = ('L','C')
	sighash_type    = 'ALL|FORKID'
	forks = [
		_finfo(478559,'000000000000000000651ef99cb9fcbe0dadde1d424bd9f15ff20136191a5eec','BTC',False)
	]
	caps = ()
	coin_amt        = 'BCHAmt'
	max_tx_fee      = '0.1'
	ignore_daemon_version = False

	def pubhash2redeem_script(self,pubhash):
		raise NotImplementedError

	def pubhash2segwitaddr(self,pubhash):
		raise NotImplementedError

class testnet(mainnet):
	addr_ver_info  = { '6f': 'p2pkh', 'c4': 'p2sh' }
	wif_ver_num    = { 'std': 'ef' }

class regtest(testnet):
	halving_interval = 150



#!/usr/bin/env python3

"""
test.unit_tests_d.ut_addrlist: address list unit tests for the MMGen suite
"""

from mmgen.color import blue
from mmgen.util import msg

from mmgen.seed import Seed
from mmgen.addr import MMGenAddrType
from mmgen.addrlist import AddrIdxList,AddrList,KeyList,KeyAddrList,ViewKeyAddrList
from mmgen.passwdlist import PasswordList
from mmgen.protocol import init_proto
from ..include.common import cfg,qmsg,vmsg

def do_test(
		list_type,
		chksum,
		idx_spec   = None,
		pw_id_str  = None,
		add_kwargs = None,
		coin       = None,
		addrtype   = None):

	qmsg(blue(f'Testing {list_type.__name__}'))
	proto = init_proto( cfg, coin or 'btc' )
	seed = Seed(cfg,seed_bin=bytes.fromhex('feedbead'*8))
	mmtype = MMGenAddrType(proto, addrtype or 'C')
	idxs = AddrIdxList(idx_spec or '1-3')

	if cfg.verbose:
		debug_addrlist_save = cfg.debug_addrlist
		cfg.debug_addrlist = True

	kwargs = {
		'seed': seed,
		'pw_idxs': idxs,
		'pw_id_str': pw_id_str,
		'pw_fmt': 'b58',
	} if pw_id_str else {
		'seed': seed,
		'addr_idxs': idxs,
		'mmtype': mmtype,
	}

	if add_kwargs:
		kwargs.update(add_kwargs)

	al = list_type( cfg, proto, **kwargs )

	al.file.format()

	qmsg(f'Filename: {al.file.filename}\n')

	vmsg(f'------------\n{al.file.fmt_data}\n------------')

	if chksum:
		assert al.chksum == chksum, f'{al.chksum} != {chksum}'

	if cfg.verbose:
		cfg.debug_addrlist = debug_addrlist_save

	return True

class unit_tests:

	altcoin_deps = ('keyaddr_xmr','viewkeyaddr')

	def idxlist(self,name,ut):
		for i,o in (
				('99,88-102,1-3,4,9,818,444-445,816',        '1-4,9,88-102,444-445,816,818'),
				('99,88-99,100,102,4-7,9,818,444-445,816,1', '1,4-7,9,88-100,102,444-445,816,818'),
				('8',             '8'),
				('2-4',           '2-4'),
				('1,2-4',         '1-4'),
				('2-4,1-9,9,1,8', '1-9'),
				('2-4,1',         '1-4'),
				('2-2',           '2'),
				('2,2',           '2'),
				('2-3',           '2-3'),
				('2,3',           '2-3'),
				('3,2',           '2-3'),
				('2,4',           '2,4'),
				('',              ''),
			):
			l = AddrIdxList(i)
			if cfg.verbose:
				msg(f'list: {list(l)}\nin:   {i}\nout:  {o}\n')
			assert l.id_str == o, f'{l.id_str} != {o}'

		return True

	def addr(self,name,ut):
		return (
			do_test(AddrList,'BCE8 082C 0973 A525','1-3') and
			do_test(AddrList,'88FA B04B A380 C1CB','199999,99-101,77-78,7,3,2-9')
		)

	def key(self,name,ut):
		return do_test(KeyList,None)

	def keyaddr(self,name,ut):
		return do_test(KeyAddrList,'4A36 AA65 8C2B 7C35')

	def keyaddr_xmr(self,name,ut):
		return do_test(KeyAddrList,'AAA2 BA69 17FC 9A88',coin='XMR',addrtype='M')

	def viewkeyaddr(self,name,ut):
		return do_test(ViewKeyAddrList,'C122 2E58 DC28 D6AE',coin='XMR',addrtype='M')

	def passwd(self,name,ut):
		return do_test(PasswordList,'FF4A B716 4513 8F8F',pw_id_str='foo')

	def passwd_bip39(self,name,ut):
		return do_test(PasswordList,'C3A8 B2B2 1AA1 FB40',pw_id_str='foo',add_kwargs={'pw_fmt':'bip39'})



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
opts: MMGen-specific command-line options processing after generic processing by share.Opts
"""
import sys,os

from .share import Opts
from .cfg import gc

def opt_preproc_debug(po):
	d = (
		('Cmdline',            ' '.join(sys.argv), False),
		('Filtered opts',      po.filtered_opts,   False),
		('User-selected opts', po.user_opts,       False),
		('Cmd args',           po.cmd_args,        False),
		('Opts',               po.opts,            True),
	)
	from .util import Msg,fmt_list
	Msg('\n=== opts.py debug ===')
	for label,data,pretty in d:
		Msg('    {:<20}: {}'.format(label,'\n' + fmt_list(data,fmt='col',indent=' '*8) if pretty else data))

long_opts_data = {
	'text': """
--, --accept-defaults      Accept defaults at all prompts
--, --coin=c               Choose coin unit. Default: BTC. Current choice: {cu_dfl}
--, --token=t              Specify an ERC20 token by address or symbol
--, --color=0|1            Disable or enable color output (enabled by default)
--, --columns=N            Force N columns of output with certain commands
--, --scroll               Use the curses-like scrolling interface for
                           tracking wallet views
--, --force-256-color      Force 256-color output when color is enabled
--, --pager                Pipe output of certain commands to pager (WIP)
--, --data-dir=path        Specify {pnm} data directory location
--, --daemon-data-dir=path Specify coin daemon data directory location
--, --daemon-id=ID         Specify the coin daemon ID
--, --ignore-daemon-version Ignore coin daemon version check
--, --http-timeout=t       Set HTTP timeout in seconds for JSON-RPC connections
--, --no-license           Suppress the GPL license prompt
--, --rpc-host=HOST        Communicate with coin daemon running on host HOST
--, --rpc-port=PORT        Communicate with coin daemon listening on port PORT
--, --rpc-user=USER        Authenticate to coin daemon using username USER
--, --rpc-password=PASS    Authenticate to coin daemon using password PASS
--, --rpc-backend=backend  Use backend 'backend' for JSON-RPC communications
--, --aiohttp-rpc-queue-len=N Use N simultaneous RPC connections with aiohttp
--, --regtest=0|1          Disable or enable regtest mode
--, --testnet=0|1          Disable or enable testnet
--, --tw-name=NAME         Specify alternate name for the BTC/LTC/BCH tracking
                         wallet (default: ‘{tw_name}’)
--, --skip-cfg-file        Skip reading the configuration file
--, --version              Print version information and exit
--, --bob                  Specify user “Bob” in MMGen regtest mode
--, --alice                Specify user “Alice” in MMGen regtest mode
--, --carol                Specify user “Carol” in MMGen regtest mode
	""",
	'code': lambda proto,help_notes,s: s.format(
			pnm    = gc.proj_name,
			cu_dfl = proto.coin,
			tw_name = help_notes('dfl_twname')
		)
}

opts_data_dfl = {
	'text': {
		'desc': '',
		'usage':'[options]',
		'options': """
-h, --help         Print this help message
--, --longhelp     Print help message for long (common) options
"""
	}
}

class UserOpts:

	def __init__(
			self,
			cfg,
			opts_data   = None,
			init_opts   = None, # dict containing opts to pre-initialize
			opt_filter  = None, # whitelist of opt letters; all others are skipped
			parse_only  = False,
			parsed_opts = None ):

		self.opts_data = od = opts_data or opts_data_dfl
		self.opt_filter = opt_filter

		od['text']['long_options'] = long_opts_data['text']

		# Make this available to usage()
		self.usage_data = od['text'].get('usage2') or od['text']['usage']

		# po: (user_opts,cmd_args,opts,filtered_opts)
		po = parsed_opts or Opts.parse_opts(od,opt_filter=opt_filter)

		cfg._args = po.cmd_args
		cfg._uopts = uopts = po.user_opts

		if init_opts: # initialize user opts to given value
			for uopt,val in init_opts.items():
				if uopt not in uopts:
					uopts[uopt] = val

		cfg._opts = self
		cfg._parsed_opts = po
		cfg._use_env = True
		cfg._use_cfg_file = not 'skip_cfg_file' in uopts

		if os.getenv('MMGEN_DEBUG_OPTS'):
			opt_preproc_debug(po)

		if 'version' in uopts:
			self.version() # exits

		if 'show_hash_presets' in uopts:
			self.show_hash_presets() # exits

		if parse_only:
			return

	def init_bottom(self,cfg):

		# print help screen only after globals initialized and locked:
		if cfg.help or cfg.longhelp:
			self.print_help(cfg) # exits

		# delete unneeded data:
		for k in ('text','notes','code'):
			if k in self.opts_data:
				del self.opts_data[k]
		del Opts.make_help
		del Opts.process_uopts
		del Opts.parse_opts

	def usage(self):
		from .util import Die
		Die(1,Opts.make_usage_str(gc.prog_name,'user',self.usage_data))

	def version(self):
		from .util import Die,fmt
		Die(0,fmt(f"""
			{gc.prog_name.upper()} version {gc.version}
			Part of {gc.proj_name} Wallet, an online/offline cryptocurrency wallet for the
			command line. Copyright (C){gc.Cdates} {gc.author} {gc.email}
		""",indent='  ').rstrip())

	def print_help(self,cfg):

		if not 'code' in self.opts_data:
			self.opts_data['code'] = {}

		from .protocol import init_proto_from_cfg
		proto = init_proto_from_cfg(cfg,need_amt=True)

		if getattr(cfg,'longhelp',None):
			self.opts_data['code']['long_options'] = long_opts_data['code']
			def remove_unneeded_long_opts():
				d = self.opts_data['text']['long_options']
				if proto.base_proto != 'Ethereum':
					d = '\n'.join(''+i for i in d.split('\n') if not '--token' in i)
				self.opts_data['text']['long_options'] = d
			remove_unneeded_long_opts()

		from .ui import do_pager
		do_pager(Opts.make_help( cfg, proto, self.opts_data, self.opt_filter ))

		sys.exit(0)

	def show_hash_presets(self):
		fs = '      {:<6} {:<3} {:<2} {}'
		from .util import msg
		from .crypto import Crypto
		msg('  Available parameters for scrypt.hash():')
		msg(fs.format('Preset','N','r','p'))
		for i in sorted(Crypto.hash_presets.keys()):
			msg(fs.format(i,*Crypto.hash_presets[i]))
		msg('  N = memory usage (power of two)\n  p = iterations (rounds)')
		sys.exit(0)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.info: Ethereum transaction info class
"""

from ....tx.info import TxInfo
from ....util import fmt,pp_fmt
from ....color import pink,yellow,blue
from ....addr import MMGenID
from ....obj import Str

class TxInfo(TxInfo):
	txinfo_hdr_fs = 'TRANSACTION DATA\n\nID={i} ({a} {c}) Sig={s} Locktime={l}\n'
	txinfo_hdr_fs_short = 'TX {i} ({a} {c}) Sig={s} Locktime={l}\n'
	txinfo_ftr_fs = fmt("""
		Total in account:  {i} {d}
		Total to spend:    {o} {d}
		Remaining balance: {C} {d}
		TX fee:            {a} {c}{r}
	""")
	fmt_keys = ('from','to','amt','nonce')

	def format_body(self,blockcount,nonmm_str,max_mmwid,enl,terse,sort):
		tx = self.tx
		m = {}
		for k in ('inputs','outputs'):
			if len(getattr(tx,k)):
				m[k] = getattr(tx,k)[0].mmid if len(getattr(tx,k)) else ''
				m[k] = ' ' + m[k].hl() if m[k] else ' ' + MMGenID.hlc(nonmm_str)
		fs = """From:      {}{f_mmid}
				To:        {}{t_mmid}
				Amount:    {} {c}
				Gas price: {g} Gwei
				Start gas: {G} Kwei
				Nonce:     {}
				Data:      {d}
				\n""".replace('\t','')
		t = tx.txobj
		td = t['data']
		return fs.format(
			*((t[k] if t[k] != '' else Str('None')).hl() for k in self.fmt_keys),
			d      = '{}... ({} bytes)'.format(td[:40],len(td)//2) if len(td) else Str('None'),
			c      = tx.proto.dcoin if len(tx.outputs) else '',
			g      = yellow(str(t['gasPrice'].to_unit('Gwei',show_decimal=True))),
			G      = yellow(str(t['startGas'].to_unit('Kwei'))),
			t_mmid = m['outputs'] if len(tx.outputs) else '',
			f_mmid = m['inputs'] )

	def format_abs_fee(self,color,iwidth):
		return self.tx.fee.fmt(color=color,iwidth=iwidth) + (' (max)' if self.tx.txobj['data'] else '')

	def format_rel_fee(self):
		return ' ({} of spend amount)'.format(
			pink('{:0.6f}%'.format( self.tx.fee / self.tx.send_amt * 100 ))
		)

	def format_verbose_footer(self):
		if self.tx.txobj['data']:
			from ..contract import parse_abi
			return '\nParsed contract data: ' + pp_fmt(parse_abi(self.tx.txobj['data']))
		else:
			return ''

class TokenTxInfo(TxInfo):
	fmt_keys = ('from','token_to','amt','nonce')

	def format_rel_fee(self):
		return ''

	def format_body(self,*args,**kwargs):
		return 'Token:     {d} {c}\n{r}'.format(
			d = self.tx.txobj['token_addr'].hl(),
			c = blue('(' + self.tx.proto.dcoin + ')'),
			r = super().format_body(*args,**kwargs ))
